from types import NoneType
from typing import NamedTuple, Union

from astronomix._physics_modules._cnn_mhd_corrector._cnn_mhd_corrector_options import (
    CNNMHDconfig,
)
from astronomix._physics_modules._cooling.cooling_options import CoolingConfig
from astronomix._physics_modules._cosmic_rays.cosmic_ray_options import CosmicRayConfig
from astronomix._physics_modules._neural_net_force._neural_net_force_options import (
    NeuralNetForceConfig,
)
from astronomix._physics_modules._stellar_wind.stellar_wind_options import WindConfig

from jaxtyping import Array, Float

from astronomix._physics_modules._turbulent_forcing._turbulent_forcing_options import TurbulentForcingConfig

# ===================== constant definition =====================

# solver modes
FINITE_VOLUME = 0
FINITE_DIFFERENCE = 1

# differentiation modes
FORWARDS = 0
BACKWARDS = 1

# limiter types
MINMOD = 0
OSHER = 1
DOUBLE_MINMOD = 2
SUPERBEE = 3
VAN_ALBADA = 4
VAN_ALBADA_PP = 5

# splitting modes
UNSPLIT = 0
SPLIT = 1

# Riemann solvers
HLL = 0
HLLC = 1
HLLC_LM = 2
LAX_FRIEDRICHS = 3
HYBRID_HLLC = 4
AM_HLLC = 5

# time integrators
# currently only for finite volume
RK2_SSP = 0
MUSCL = 1
# currently only for finite difference
RK4_SSP = 2

# boundary conditions
OPEN_BOUNDARY = 0
REFLECTIVE_BOUNDARY = 1
PERIODIC_BOUNDARY = 2
MHD_JET_BOUNDARY = 3

GAS_STATE = 0
VELOCITY_ONLY = 1
MAGNETIC_FIELD_ONLY = 2

# geometry types
CARTESIAN = 0
CYLINDRICAL = 1
SPHERICAL = 2

# axes
VARAXIS = 0
XAXIS = 1
YAXIS = 2
ZAXIS = 3

# boundary handling modes
GHOST_CELLS = 0
PERIODIC_ROLL = 1

# self-gravity versions
SIMPLE_SOURCE_TERM = 0
DONOR_ACCOUNTING = 1
RIEMANN_SPLIT = 2
RIEMANN_SPLIT_UNSTABLE = 3
HALF_SPLIT = 4

# Magnetic part integrators for split MHD
IMPLICIT_MIDPOINT = 0
IMPLICIT_EULER = 1

# Numerical precision
SINGLE_PRECISION = 0
DOUBLE_PRECISION = 1

# ============================================================

# ===================== type definitions =====================

STATE_TYPE = Union[
    Float[Array, "num_vars num_cells_x"],
    Float[Array, "num_vars num_cells_x num_cells_y"],
    Float[Array, "num_vars num_cells_x num_cells_y num_cells_z"],
]

STATE_TYPE_ALTERED = Union[
    Float[Array, "num_vars num_cells_a"],
    Float[Array, "num_vars num_cells_a num_cells_b"],
    Float[Array, "num_vars num_cells_a num_cells_b num_cells_c"],
]

FIELD_TYPE = Union[
    Float[Array, "num_cells_x"],
    Float[Array, "num_cells_x num_cells_y"],
    Float[Array, "num_cells_x num_cells_y num_cells_z"],
]

class SnapshotSettings(NamedTuple):
    """Settings for the snapshot output of the simulation."""

    #: Whether to return states during the simulation.
    return_states: bool = True

    #: Whether to return the final state of the simulation.
    return_final_state: bool = False

    #: Whether to return the total mass at the times the snapshots were taken.
    return_total_mass: bool = False

    #: Whether to return the total energy at the times the snapshots were taken.
    return_total_energy: bool = False

    #: Whether to return internal energy
    return_internal_energy: bool = False

    #: Whether to return kinetic energy
    return_kinetic_energy: bool = False

    #: Whether to return gravitational energy
    return_gravitational_energy: bool = False

    #: Whether to return radial momentum
    return_radial_momentum: bool = False

    #: Whether to return the magnetic field divergence
    #: NOTE: currently only implemented for finite difference MHD
    return_magnetic_divergence: bool = False


class BoundarySettings1D(NamedTuple):
    left_boundary: int = OPEN_BOUNDARY
    right_boundary: int = OPEN_BOUNDARY


class BoundarySettings(NamedTuple):
    x: BoundarySettings1D = BoundarySettings1D()
    y: BoundarySettings1D = BoundarySettings1D()
    z: BoundarySettings1D = BoundarySettings1D()


class SimulationConfig(NamedTuple):
    """
    Configuration object for the simulation.
    The simulation configuration are parameters defining
    the simulation where changes necessitate recompilation.
    """

    # Static simulation parameters

    #: Basic solver mode, either finite volume or finite difference.
    #: FINITE_DIFFERENCE is for now only planned for the HOW_MHD
    #: scheme (Jeongbhin Seo, Dongsu Ryu, 2023).
    solver_mode: int = FINITE_VOLUME

    #: Precision mode.
    numerical_precision: int = SINGLE_PRECISION

    #: Debug runtime errors, throws exceptions
    #: on e.g. negative pressure or density.
    #: Significantly reduces performance.
    runtime_debugging: bool = False

    #: Donate the state arrays to the time integration function
    #: to reduce memory allocations. If activated, the
    #: initial state arrays will be invalid after
    #: the simulation.
    donate_state: bool = False

    #: Memory analysis of the main time integration
    #: function
    memory_analysis: bool = False

    #: Print the elapsed time of the simulation
    print_elapsed_time: bool = False

    #: Activate progress bar
    progress_bar: bool = False

    #: The number of dimensions of the simulation.
    dimensionality: int = 1

    #: Use a struct for the state.
    state_struct: bool = False

    #: The geometry of the simulation.
    geometry: int = CARTESIAN

    #: Magnetohydrodynamics switch.
    mhd: bool = False

    #: Integrator used for the magnetic part in the FV MHD scheme.
    fv_magnetic_integrator: int = IMPLICIT_MIDPOINT

    #: Enforce positivity of density and pressure.
    #: NOTE: CURRENTLY ONLY IMPLEMENTED FOR 
    #: FINITE DIFFERENCE MODE.
    enforce_positivity: bool = True

    #: Self gravity switch, currently only
    #: for periodic boundaries.
    self_gravity: bool = False
    self_gravity_version: int = DONOR_ACCOUNTING

    #: The size of the simulation box.
    box_size: float = 1.0

    #: The number of cells in the simulation (including ghost cells).
    num_cells: int = 400

    #: The reconstruction order is the number of
    #: cells on each side of the cell of interest
    #: used to calculate the gradients for the
    #: reconstruction at the interfaces.
    reconstruction_order: int = 1

    #: The limiter for the reconstruction.
    #: Only affects finite volume mode.
    limiter: int = MINMOD

    #: The Riemann solver used
    #: Only for finite volume mode.
    riemann_solver: int = HLL

    #: Dimensional splitting / unsplit mode.
    #: Note that the UNSPLIT scheme currently
    #: interferes with energy conservation in settings
    #: with self-gravity.
    split: int = UNSPLIT

    #: Time integration method.
    time_integrator: int = RK2_SSP

    # Explanation of the ghost cells
    #                                |---------|
    #                           |---------|
    # stencil              |---------|
    # cells            || 1g | 2g | 3c | 4g | 5g ||
    # reconstructions        |L  R|L  R|L  R|    |
    # fluxes                     -->  -->
    # update                      | 3c'|
    # --> all others are ghost cells

    #: The number of ghost cells.
    num_ghost_cells: int = reconstruction_order + 1

    #: Grid spacing.
    grid_spacing: float = box_size / num_cells

    #: Explicit boundary handling mode.
    boundary_handling: int = GHOST_CELLS

    #: Boundary settings for the simulation.
    boundary_settings: Union[NoneType, BoundarySettings1D, BoundarySettings] = None

    #: Enables a fixed timestep for the simulation
    #: based on the specified number of timesteps.
    fixed_timestep: bool = False

    #: Exactly reach the end time. In adaptive timestepping,
    #: one might otherwise overshoot.
    exact_end_time: bool = True

    #: Adds the sources with the current timestep to
    #: a hypothetical state to estimate the actual timestep.
    #: Useful for time-dependent sources, but additional
    #: computational overhead.
    source_term_aware_timestep: bool = False

    #: The number of timesteps for the fixed timestep mode.
    num_timesteps: int = 1000

    #: Use a maximum timestep in adaptive timestep mode.
    use_max_adaptive_timestep: bool = True

    #: The differentiation mode one whats to use
    #: the solver in (forwards or backwards).
    differentiation_mode: int = FORWARDS

    #: The number of checkpoints used in the setup
    #: with backwards differetiability and adaptive
    #: time stepping.
    num_checkpoints: int = 100

    #: Return intermediate snapshots of the time evolution
    #: instead of only the final fluid state.
    return_snapshots: bool = False

    #: Snapshot settings
    snapshot_settings: SnapshotSettings = SnapshotSettings()

    #: Call a user given function on the snapshot data,
    #: e.g. for saving or plotting. Must have signature
    #: callback(time, state, registered_variables).
    activate_snapshot_callback: bool = False

    #: Return snapshots at specific time points.
    use_specific_snapshot_timepoints: bool = False

    #: The number of snapshots to return.
    num_snapshots: int = 10

    #: Fallback to the first order Godunov scheme.
    first_order_fallback: bool = False

    # physical modules

    #: Turbulent forcing configuration.
    turbulent_forcing_config: TurbulentForcingConfig = TurbulentForcingConfig()

    #: The configuration for the stellar wind module.
    wind_config: WindConfig = WindConfig()

    #: Cosmic rays
    cosmic_ray_config: CosmicRayConfig = CosmicRayConfig()

    #: The configuration for the cooling module.
    cooling_config: CoolingConfig = CoolingConfig()

    #: Configuration of the neural network force module.
    neural_net_force_config: NeuralNetForceConfig = NeuralNetForceConfig()

    #: Configuration of the CNN MHD corrector module.
    cnn_mhd_corrector_config: CNNMHDconfig = CNNMHDconfig()


def finalize_config(config: SimulationConfig, state_shape) -> SimulationConfig:
    """Finalizes the simulation configuration."""

    num_cells = state_shape[-1]
    config = config._replace(num_cells=num_cells)

    if config.dimensionality == 2:
        num_cells_x, num_cells_y = state_shape[-2:]
        if num_cells_x != num_cells_y:
            raise ValueError("The number of cells in x and y must be equal.")
    elif config.dimensionality == 3:
        num_cells_x, num_cells_y, num_cells_z = state_shape[-3:]
        if num_cells_x != num_cells_y or num_cells_x != num_cells_z:
            raise ValueError("The number of cells in x, y and z must be equal.")

    config = config._replace(grid_spacing=config.box_size / config.num_cells)

    if config.geometry == SPHERICAL:
        print(
            "For spherical geometry, only HLL is currently supported. Also, only the unsplit mode has been tested."
        )
        config = config._replace(grid_spacing=config.box_size / config.num_cells)

        if config.riemann_solver != HLL:
            print("Setting HLL Riemann solver for spherical geometry.")
            config = config._replace(riemann_solver=HLL)

        if config.split != SPLIT:
            print("Setting unsplit mode for spherical geometry")
            config = config._replace(split=SPLIT)

        if config.limiter == VAN_ALBADA or config.limiter == VAN_ALBADA_PP:
            print("Setting minmod limiter for spherical geometry")
            config = config._replace(limiter=MINMOD)

        if config.time_integrator != MUSCL:
            print("Setting MUSCL time integrator for spherical geometry")
            config = config._replace(time_integrator=MUSCL)

    if config.self_gravity and (config.limiter != MINMOD):
        print(
            "Curiously, in self-gravitating systems, the VAN_ALBADA limiters seem to cause crashes."
        )
        print("Setting DOUBLE_MINMOD limiter for self-gravity.")
        config = config._replace(limiter=MINMOD)

    # finite difference specific checks
    if config.solver_mode == FINITE_DIFFERENCE:
        if not config.mhd:
            raise ValueError(
                "Finite difference solver mode is currently " \
                "only supported for MHD simulations. This will be easy to extend, " \
                "feel free to contribute."
            )

        if config.dimensionality != 3:
            raise ValueError(
                "Finite difference solver mode is currently " \
                "only supported for 3D simulations. This will be easy to extend, " \
                "feel free to contribute."
            )
        
        if config.boundary_settings != BoundarySettings(
            BoundarySettings1D(
                left_boundary=PERIODIC_BOUNDARY, right_boundary=PERIODIC_BOUNDARY
            ),
            BoundarySettings1D(
                left_boundary=PERIODIC_BOUNDARY, right_boundary=PERIODIC_BOUNDARY
            ),
            BoundarySettings1D(
                left_boundary=PERIODIC_BOUNDARY, right_boundary=PERIODIC_BOUNDARY
            ),
        ):
            raise ValueError(
                "Finite difference solver mode currently only supports periodic boundaries."
            )
        
        if config.time_integrator != RK4_SSP:
            print(
                "Setting time integrator to RK4_SSP for finite difference solver mode."
            )
            config = config._replace(time_integrator=RK4_SSP)
        
        if config.boundary_handling != PERIODIC_ROLL:
            print(
                "Setting boundary handling to " \
                "PERIODIC_ROLL for finite difference solver mode."
            )
            config = config._replace(boundary_handling=PERIODIC_ROLL)

    # set boundary conditions if not set
    if config.boundary_settings is None:
        if config.geometry == CARTESIAN:
            print("Automatically setting open boundaries for Cartesian geometry.")
            if config.dimensionality == 1:
                config = config._replace(
                    boundary_settings=BoundarySettings1D(
                        left_boundary=OPEN_BOUNDARY, right_boundary=OPEN_BOUNDARY
                    )
                )
            else:
                config = config._replace(boundary_settings=BoundarySettings())
        elif config.geometry == SPHERICAL and config.dimensionality == 1:
            print(
                "Automatically setting reflective left and open right boundary for spherical geometry."
            )
            config = config._replace(
                boundary_settings=BoundarySettings1D(
                    left_boundary=REFLECTIVE_BOUNDARY, right_boundary=OPEN_BOUNDARY
                )
            )

    if config.wind_config.stellar_wind:
        print(
            "For stellar wind simulations, we need source term aware timesteps, turning on."
        )
        config = config._replace(source_term_aware_timestep=True)

    if (
        config.self_gravity
        and (config.riemann_solver == HLLC or config.riemann_solver == HLLC_LM)
        and config.riemann_solver != RIEMANN_SPLIT
    ):
        print("Consider using RIEMANN_SPLIT as the self_gravity_version.")

    return config



# general imports
import jax.numpy as jnp
import jax
from functools import partial

# runtime debugging
from jax.experimental import checkify

# type checking imports
from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker
from typing import Union

# general astronomix imports
from astronomix._finite_volume._riemann_solver._riemann_solver import _riemann_solver
from astronomix._finite_volume._magnetic_update._magnetic_field_update import magnetic_update
from astronomix._physics_modules._self_gravity._self_gravity import _apply_self_gravity
from astronomix._stencil_operations._stencil_operations import _stencil_add
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import (
    CARTESIAN,
    GHOST_CELLS,
    RK2_SSP,
    SPHERICAL,
    STATE_TYPE,
    UNSPLIT,
    VAN_ALBADA_PP,
    SimulationConfig,
)

from astronomix._geometry.geometric_terms import _pressure_nozzling_source
from astronomix._finite_volume._state_evolution.reconstruction import (
    _reconstruct_at_interface_split,
    _reconstruct_at_interface_unsplit,
    _reconstruct_at_interface_unsplit_single,
)
from astronomix._geometry.boundaries import _boundary_handler
from astronomix._fluid_equations._equations import (
    primitive_state_from_conserved,
    conserved_state_from_primitive,
)
from astronomix._finite_volume._riemann_solver._lax_friedrichs import _lax_friedrichs_solver
from astronomix.option_classes.simulation_params import SimulationParams

# -------------------------------------------------------------
# ====================== ↓ Split Scheme ↓ =====================
# -------------------------------------------------------------


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables", "axis"])
def _evolve_state_along_axis(
    primitive_state: STATE_TYPE,
    grid_spacing: Union[float, Float[Array, ""]],
    dt: Float[Array, ""],
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
    axis: int,
) -> STATE_TYPE:
    
    if config.boundary_handling == GHOST_CELLS:
        primitive_state = _boundary_handler(primitive_state, config)

    # get conserved variables
    conservative_states = conserved_state_from_primitive(
        primitive_state, gamma, config, registered_variables
    )

    if config.first_order_fallback:
        primitive_state_left = jnp.roll(primitive_state, shift=1, axis=axis)
        primitive_state_right = primitive_state
    else:
        primitive_state_left, primitive_state_right = _reconstruct_at_interface_split(
            primitive_state, dt, gamma, config, helper_data, registered_variables, axis
        )

    fluxes = _riemann_solver(
        primitive_state_left,
        primitive_state_right,
        primitive_state,
        gamma,
        config,
        registered_variables,
        axis,
    )

    # ================ update the conserved variables =================

    # usual cartesian case
    if config.geometry == CARTESIAN:
        conserved_change = (
            1
            / grid_spacing
            * _stencil_add(fluxes, indices=(0, 1), factors=(1.0, -1.0), axis=axis)
            * dt
        )

    # in spherical geometry, we have to take special care
    elif config.geometry == SPHERICAL and config.dimensionality == 1 and axis == 1:
        r = helper_data.geometric_centers
        r_hat_alpha = helper_data.r_hat_alpha

        alpha = config.geometry

        r_plus_half = r + grid_spacing / 2
        r_minus_half = r - grid_spacing / 2

        # calculate the source terms
        nozzling_source = _pressure_nozzling_source(
            primitive_state, config, helper_data, registered_variables
        )

        # update the conserved variables using the fluxes and source terms
        conserved_change = (
            1
            / r_hat_alpha
            * (
                +(
                    r_minus_half**alpha * fluxes
                    - r_plus_half**alpha * jnp.roll(fluxes, shift=-1, axis=axis)
                )
                / grid_spacing
                + nozzling_source
            )
            * dt
        )

    # misconfiguration
    else:
        raise ValueError("Geometry and dimensionality combination not supported.")

    # =================================================================

    conservative_states = conservative_states + conserved_change

    primitive_state = primitive_state_from_conserved(
        conservative_states, gamma, config, registered_variables
    )
    if config.boundary_handling == GHOST_CELLS:
        primitive_state = _boundary_handler(primitive_state, config)

    # check if the pressure is still positive
    p = primitive_state[registered_variables.pressure_index]
    rho = primitive_state[registered_variables.density_index]

    if config.runtime_debugging:
        checkify.check(
            jnp.all(p >= 0),
            "pressure needs to be non-negative, minimum pressure {pmin} at index {index}",
            pmin=jnp.min(p),
            index=jnp.unravel_index(jnp.argmin(p), p.shape),
        )
        checkify.check(
            jnp.all(rho >= 0),
            "density needs to be non-negative, minimum density {rhomin} at index {index}",
            rhomin=jnp.min(rho),
            index=jnp.unravel_index(jnp.argmin(rho), rho.shape),
        )

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _evolve_gas_state_split(
    primitive_state: STATE_TYPE,
    dt: Float[Array, ""],
    gamma: Union[float, Float[Array, ""]],
    gravitational_constant: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    params: SimulationParams,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    if config.dimensionality == 1:
        if config.self_gravity:
            old_primitive_state = primitive_state

        primitive_state = _evolve_state_along_axis(
            primitive_state,
            config.grid_spacing,
            dt,
            gamma,
            config,
            helper_data,
            registered_variables,
            1,
        )

        if config.self_gravity:
            primitive_state = _apply_self_gravity(
                primitive_state,
                old_primitive_state,
                config,
                params,
                registered_variables,
                helper_data,
                gamma,
                gravitational_constant,
                dt,
            )

    elif config.dimensionality == 2:
        if config.self_gravity:
            old_primitive_state = primitive_state

        primitive_state = _evolve_state_along_axis(
            primitive_state,
            config.grid_spacing,
            dt / 2,
            gamma,
            config,
            helper_data,
            registered_variables,
            1,
        )
        primitive_state = _evolve_state_along_axis(
            primitive_state,
            config.grid_spacing,
            dt,
            gamma,
            config,
            helper_data,
            registered_variables,
            2,
        )
        primitive_state = _evolve_state_along_axis(
            primitive_state,
            config.grid_spacing,
            dt / 2,
            gamma,
            config,
            helper_data,
            registered_variables,
            1,
        )

        if config.self_gravity:
            primitive_state = _apply_self_gravity(
                primitive_state,
                old_primitive_state,
                config,
                params,
                registered_variables,
                helper_data,
                gamma,
                gravitational_constant,
                dt,
            )

    elif config.dimensionality == 3:
        if config.self_gravity:
            old_primitive_state = primitive_state

        primitive_state = _evolve_state_along_axis(
            primitive_state,
            config.grid_spacing,
            dt / 2,
            gamma,
            config,
            helper_data,
            registered_variables,
            1,
        )
        primitive_state = _evolve_state_along_axis(
            primitive_state,
            config.grid_spacing,
            dt / 2,
            gamma,
            config,
            helper_data,
            registered_variables,
            2,
        )
        primitive_state = _evolve_state_along_axis(
            primitive_state,
            config.grid_spacing,
            dt,
            gamma,
            config,
            helper_data,
            registered_variables,
            3,
        )
        primitive_state = _evolve_state_along_axis(
            primitive_state,
            config.grid_spacing,
            dt / 2,
            gamma,
            config,
            helper_data,
            registered_variables,
            2,
        )
        primitive_state = _evolve_state_along_axis(
            primitive_state,
            config.grid_spacing,
            dt / 2,
            gamma,
            config,
            helper_data,
            registered_variables,
            1,
        )

        if config.self_gravity:
            primitive_state = _apply_self_gravity(
                primitive_state,
                old_primitive_state,
                config,
                params,
                registered_variables,
                helper_data,
                gamma,
                gravitational_constant,
                dt,
            )

    else:
        raise ValueError("Dimensionality not supported.")

    return primitive_state


# -------------------------------------------------------------
# ====================== ↑ Split Scheme ↑ =====================
# -------------------------------------------------------------

# -------------------------------------------------------------
# ===================== ↓ Unsplit Scheme ↓ ====================
# -------------------------------------------------------------


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _evolve_gas_state_unsplit_inner(
    primitive_state: STATE_TYPE,
    dt: Float[Array, ""],
    gamma: Union[float, Float[Array, ""]],
    gravitational_constant: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    params: SimulationParams,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    
    if config.boundary_handling == GHOST_CELLS:
        primitive_state = _boundary_handler(primitive_state, config)
    conservative_states = conserved_state_from_primitive(
        primitive_state, gamma, config, registered_variables
    )

    # in case of the van albada pp limiter, the limited
    # gradients along all dimensions are needed at once for
    # the proper multidimensional limiting
    if config.limiter == VAN_ALBADA_PP:
        # get left and right states along all dimensions
        pls, prs = _reconstruct_at_interface_unsplit(
            primitive_state,
            dt,
            gamma,
            config,
            params,
            helper_data,
            registered_variables
        )

    for axis in range(1, config.dimensionality + 1):

        if config.boundary_handling == GHOST_CELLS:
            primitive_state = _boundary_handler(primitive_state, config)

        if config.limiter == VAN_ALBADA_PP:
            primitives_left_interface = pls[axis - 1]
            primitives_right_interface = prs[axis - 1]
        else:
            primitives_left_interface, primitives_right_interface = (
                _reconstruct_at_interface_unsplit_single(
                    primitive_state, config, helper_data, axis
                )
            )

        # get the fluxes at the interfaces
        fluxes = _riemann_solver(
            primitives_left_interface,
            primitives_right_interface,
            primitive_state,
            gamma,
            config,
            registered_variables,
            axis,
        )

        # update the conserved variables
        conserved_change = (
            1
            / config.grid_spacing
            * _stencil_add(fluxes, indices=(0, 1), factors=(1.0, -1.0), axis=axis)
            * dt
        )
        conservative_states += conserved_change

    # update the primitive state
    primitive_state = primitive_state_from_conserved(
        conservative_states, gamma, config, registered_variables
    )

    if config.boundary_handling == GHOST_CELLS:
        primitive_state = _boundary_handler(primitive_state, config)

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _evolve_gas_state_unsplit(
    primitive_state: STATE_TYPE,
    dt: Float[Array, ""],
    gamma: Union[float, Float[Array, ""]],
    gravitational_constant: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    params: SimulationParams,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:

    if config.self_gravity:
        old_primitive_state = primitive_state

    if config.time_integrator == RK2_SSP:
        primitive_state_1 = _evolve_gas_state_unsplit_inner(
            primitive_state,
            dt,
            gamma,
            gravitational_constant,
            config,
            params,
            helper_data,
            registered_variables,
        )

        primitive_state_2 = _evolve_gas_state_unsplit_inner(
            primitive_state_1,
            dt,
            gamma,
            gravitational_constant,
            config,
            params,
            helper_data,
            registered_variables,
        )

        conserved_state = conserved_state_from_primitive(
            primitive_state, gamma, config, registered_variables
        )

        conserved_state_2 = conserved_state_from_primitive(
            primitive_state_2, gamma, config, registered_variables
        )

        # RK2
        conserved_state = 0.5 * (conserved_state + conserved_state_2)

        primitive_state = primitive_state_from_conserved(
            conserved_state, gamma, config, registered_variables
        )
    else:
        raise ValueError(
            "Only the RK2 SSP time integrator is currently supported for the unsplit scheme."
        )

    if config.self_gravity:
        primitive_state = _apply_self_gravity(
            primitive_state,
            old_primitive_state,
            config,
            params,
            registered_variables,
            helper_data,
            gamma,
            gravitational_constant,
            dt,
        )

    return primitive_state

# -------------------------------------------------------------
# ===================== ↑ Unsplit Scheme ↑ ====================
# -------------------------------------------------------------


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _evolve_state_fv(
    primitive_state: STATE_TYPE,
    dt: Float[Array, ""],
    gamma: Union[float, Float[Array, ""]],
    gravitational_constant: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    params: SimulationParams,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    if config.mhd:
        if config.dimensionality > 1:

            # HERE WE EXPLICITLY ASSUME THAT THE LAST 3 INDICES 
            # ARE THE MAGNETIC FIELD COMPONENTS
            registered_variables_gas = registered_variables._replace(
                num_vars=registered_variables.num_vars - 3
            )

            gas_state = primitive_state[:-3, ...]
            magnetic_field = primitive_state[-3:, ...]

            if config.split == UNSPLIT:
                evolved_gas = _evolve_gas_state_unsplit(
                    gas_state,
                    dt / 2,
                    gamma,
                    gravitational_constant,
                    config,
                    params,
                    helper_data,
                    registered_variables_gas,
                )
            else:
                evolved_gas = _evolve_gas_state_split(
                    gas_state,
                    dt / 2,
                    gamma,
                    gravitational_constant,
                    config,
                    params,
                    helper_data,
                    registered_variables_gas,
                )

            magnetic_field, evolved_gas = magnetic_update(
                magnetic_field,
                evolved_gas,
                config.grid_spacing,
                dt,
                registered_variables,
                config,
            )

            if config.split == UNSPLIT:
                evolved_gas = _evolve_gas_state_unsplit(
                    evolved_gas,
                    dt / 2,
                    gamma,
                    gravitational_constant,
                    config,
                    params,
                    helper_data,
                    registered_variables_gas,
                )
            else:
                evolved_gas = _evolve_gas_state_split(
                    evolved_gas,
                    dt / 2,
                    gamma,
                    gravitational_constant,
                    config,
                    params,
                    helper_data,
                    registered_variables_gas,
                )

            return jnp.concatenate((evolved_gas, magnetic_field), axis=0)
        else:
            raise ValueError("MHD currently not supported in 1D.")

    else:
        if config.split == UNSPLIT:
            return _evolve_gas_state_unsplit(
                primitive_state,
                dt,
                gamma,
                gravitational_constant,
                config,
                params,
                helper_data,
                registered_variables,
            )
        else:
            return _evolve_gas_state_split(
                primitive_state,
                dt,
                gamma,
                gravitational_constant,
                config,
                params,
                helper_data,
                registered_variables,
            )


from typing import Union
from astronomix._fluid_equations._fluxes import _euler_flux
from astronomix._fluid_equations._equations import conserved_state_from_primitive, speed_of_sound
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import STATE_TYPE, SimulationConfig


import jax
import jax.numpy as jnp
from beartype import beartype as typechecker
from jaxtyping import Array, Float, jaxtyped


from functools import partial


# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit, static_argnames=["config", "registered_variables", "flux_direction_index"]
)
def _lax_friedrichs_solver(
    primitives_left: STATE_TYPE,
    primitives_right: STATE_TYPE,
    primitive_state: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    flux_direction_index: int,
) -> STATE_TYPE:

    # the flux such that the array at position i stores the interface
    # from from i-1 to i.

    # primitives left at i is the left state at the interface
    # between i-1 and i so the right extrapolation from the cell i-1

    # primitives right at i is the right state at the interface
    # between i-1 and i so the left extrapolation from the cell i


    rho_L = primitives_left[registered_variables.density_index]
    u_L = primitives_left[flux_direction_index]

    rho_R = primitives_right[registered_variables.density_index]
    u_R = primitives_right[flux_direction_index]

    p_L = primitives_left[registered_variables.pressure_index]
    p_R = primitives_right[registered_variables.pressure_index]

    conserved_left = conserved_state_from_primitive(
        primitives_left, gamma, config, registered_variables
    )
    conserved_right = conserved_state_from_primitive(
        primitives_right, gamma, config, registered_variables
    )

    c_L = speed_of_sound(rho_L, p_L, gamma)
    c_R = speed_of_sound(rho_R, p_R, gamma)

    # alpha = jnp.max(jnp.maximum(jnp.abs(u_L) + c_L, jnp.abs(u_R) + c_R))
    u = primitive_state[flux_direction_index]
    rho = primitive_state[registered_variables.density_index]
    p = primitive_state[registered_variables.pressure_index]
    c = speed_of_sound(rho, p, gamma)
    # THE COMMON ALPHA PARAMETER MAKES NO SENSE (?) - LOOK INTO PAPER AGAIN???
    alpha = jnp.max(jnp.abs(u) + c)

    fluxes_left = _euler_flux(
        primitives_left, gamma, config, registered_variables, flux_direction_index
    )
    fluxes_right = _euler_flux(
        primitives_right, gamma, config, registered_variables, flux_direction_index
    )
    fluxes = 0.5 * (fluxes_left + fluxes_right) - 0.5 * alpha * (
        conserved_right - conserved_left
    )

    return fluxes



"""
Convenience functions for operations that combine multiple elements
of an array based on some stencil, e.g. b_i <- a_{i + 1} + a_{i - 1}.
Allows for code "closer to the math".
"""

# general
import jax
import jax.numpy as jnp
from functools import partial

# typechecking
from beartype import beartype as typechecker
from jaxtyping import Array, Float, jaxtyped
from typing import Tuple, Union


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["shift", "axis"])
def custom_roll(input_array: jnp.ndarray, shift: int, axis: int) -> jnp.ndarray:
    i = (-shift) % input_array.shape[axis]
    return jax.lax.concatenate(
        [
            jax.lax.slice_in_dim(input_array, i, input_array.shape[axis], axis=axis),
            jax.lax.slice_in_dim(input_array, 0, i, axis=axis),
        ],
        dimension=axis,
    )


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["indices", "axis"])
def _stencil_add(
    input_array: jnp.ndarray,
    indices: Tuple[int, ...],
    factors: Tuple[Union[float, Float[Array, ""]], ...],
    axis: int,
) -> jnp.ndarray:
    """
    Combines elements of an array additively
        output_i <- sum_j factors_j * input_array_{i + indices_j}

    Args:
        input_array: The array to operate on.
        indices: output_i <- sum_j factors_j * input_array_{i + indices_j}
        factors: output_i <- sum_j factors_j * input_array_{i + indices_j}
        axis: The axis along which to operate.

    Returns:
        output_i <- sum_j factors_j * input_array_{i + indices_j}
    """

    output = sum(
        factor * custom_roll(input_array, -index, axis=axis)
        for factor, index in zip(factors, indices)
    )

    return output



# general
import jax.numpy as jnp
import jax
from functools import partial

# typing
from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker
from typing import Union

# astronomix classes
from astronomix.variable_registry.registered_variables import RegisteredVariables

# NOTE: currently only supports 1d setups, TODO: generalize

# HERE SET MANUALLY, SHOULD COME FROM
# THE SIMULATION PARAMS
gamma_gas = 5 / 3
gamma_cr = 4 / 3


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["registered_variables"])
def total_energy_from_primitives_with_crs(
    primitive_state: Float[Array, "num_vars num_cells"],
    registered_variables: RegisteredVariables,
) -> Float[Array, "num_cells"]:
    """
    Calculates the total energy density from primitive variables in a system with cosmic rays.

    Args:
        primitive_state: Array of primitive variables
        registered_variables: Object containing indices for accessing different physical quantities

    Returns:
        Total energy density array
    """

    # get the cosmic ray pressure
    cosmic_ray_pressure = (
        primitive_state[registered_variables.cosmic_ray_n_index] ** gamma_cr
    )

    # get the cosmic ray energy (density)
    cosmic_ray_energy = cosmic_ray_pressure / (gamma_cr - 1)

    # get the gas pressure
    gas_pressure = (
        primitive_state[registered_variables.pressure_index] - cosmic_ray_pressure
    )

    # get the gas energy
    rho_gas = primitive_state[registered_variables.density_index]
    velocity = primitive_state[registered_variables.velocity_index]
    gas_energy = gas_pressure / (gamma_gas - 1) + 0.5 * rho_gas * velocity**2

    # total energy
    E_tot = gas_energy + cosmic_ray_energy

    return E_tot


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["registered_variables"])
def gas_pressure_from_primitives_with_crs(
    primitive_state: Float[Array, "num_vars num_cells"],
    registered_variables: RegisteredVariables,
) -> Float[Array, "num_cells"]:
    """
    Calculates the gas pressure from the primitive state when cosmic rays
    are considered in the simulation.

    Args:
        primitive_state: Array of primitive variables
        registered_variables: Object containing indices for accessing different physical quantities

    Returns:
        gas pressure
    """

    # get the cosmic ray pressure
    cosmic_ray_pressure = (
        primitive_state[registered_variables.cosmic_ray_n_index] ** gamma_cr
    )

    # return the gas pressure
    return primitive_state[registered_variables.pressure_index] - cosmic_ray_pressure


# TODO: make 2D and 3D ready
# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["registered_variables"])
def total_pressure_from_conserved_with_crs(
    conserved_state: Float[Array, "num_vars num_cells"],
    registered_variables: RegisteredVariables,
) -> Float[Array, "num_cells"]:
    """
    Calculates the total pressure from the conserved state when cosmic rays
    are considered in the simulation.

    Args:
        primitive_state: Array of primitive variables
        registered_variables: Object containing indices for accessing different physical quantities

    Returns:
        total pressure
    """

    # get the cosmic ray pressure
    cosmic_ray_pressure = (
        conserved_state[registered_variables.cosmic_ray_n_index] ** gamma_cr
    )

    # get the cosmic ray energy (density)
    cosmic_ray_energy = cosmic_ray_pressure / (gamma_cr - 1)

    # get the gas energy
    gas_energy = (
        conserved_state[registered_variables.pressure_index] - cosmic_ray_energy
    )

    # get the gas pressure
    rho_gas = conserved_state[registered_variables.density_index]
    velocity = conserved_state[registered_variables.velocity_index] / rho_gas
    gas_pressure = (gas_energy - 0.5 * rho_gas * velocity**2) * (gamma_gas - 1)

    # get the total pressure
    total_pressure = cosmic_ray_pressure + gas_pressure

    return total_pressure


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["registered_variables"])
def speed_of_sound_crs(
    primitive_state: Float[Array, "num_vars num_cells"],
    registered_variables: RegisteredVariables,
) -> Float[Array, "num_cells"]:
    """
    Calculates the speed of sound from the primitive state
    when cosmic rays are considered in the simulation, where
    c_s = sqrt((gamma_gas * P_gas + gamma_cr * P_CR) / rho)

    Args:
        primitive_state: Array of primitive variables
        registered_variables: Object containing indices for accessing different physical quantities

    Returns:
        sound speed
    """

    # get the cosmic ray pressure
    cosmic_ray_pressure = (
        primitive_state[registered_variables.cosmic_ray_n_index] ** gamma_cr
    )

    # get the gas pressure
    gas_pressure = (
        primitive_state[registered_variables.pressure_index] - cosmic_ray_pressure
    )

    return jnp.sqrt(
        (gamma_gas * gas_pressure + gamma_cr * cosmic_ray_pressure)
        / primitive_state[registered_variables.density_index]
    )



import jax.numpy as jnp
import jax
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix._fluid_equations._equations import (
    conserved_state_from_primitive,
    pressure_from_energy,
    primitive_state_from_conserved,
)

from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker

from functools import partial

from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import STATE_TYPE, SimulationConfig
from astronomix.option_classes.simulation_params import SimulationParams

from astronomix._physics_modules._stellar_wind.stellar_wind_options import WindParams
from astronomix._physics_modules._stellar_wind.stellar_wind_options import (
    WindConfig,
    MEO,
    MEI,
    EI,
)

from typing import Union


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _wind_injection(
    primitive_state: STATE_TYPE,
    dt: Float[Array, ""],
    config: SimulationConfig,
    params: SimulationParams,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    """Inject stellar wind into the simulation.

    Args:
        primitive_state: The primitive state array.
        dt: The time step.
        config: The simulation configuration.
        params: The simulation parameters.
        helper_data: The helper data.

    Returns:
        The primitive state array with the stellar wind injected.
    """

    if config.dimensionality == 1:
        if config.wind_config.wind_injection_scheme == MEO:
            primitive_state = _wind_meo(
                params.wind_params,
                primitive_state,
                dt,
                helper_data,
                config.num_ghost_cells,
                config.wind_config.num_injection_cells,
                params.gamma,
            )
        elif config.wind_config.wind_injection_scheme == MEI:
            primitive_state = _wind_mei(
                params.wind_params,
                primitive_state,
                dt,
                config,
                helper_data,
                config.num_ghost_cells,
                config.wind_config.num_injection_cells,
                params.gamma,
                registered_variables,
            )
        elif config.wind_config.wind_injection_scheme == EI:
            primitive_state = _wind_ei(
                params.wind_params,
                primitive_state,
                dt,
                helper_data,
                config.num_ghost_cells,
                config.wind_config.num_injection_cells,
                params.gamma,
                registered_variables,
            )
        else:
            raise ValueError("Invalid wind injection scheme")
    elif config.dimensionality == 3:
        if config.wind_config.wind_injection_scheme == EI:
            primitive_state = _wind_ei3D(
                params.wind_params,
                primitive_state,
                dt,
                config,
                helper_data,
                config.num_ghost_cells,
                config.wind_config.num_injection_cells,
                params.gamma,
                registered_variables,
            )
        else:
            raise ValueError("Invalid wind injection scheme")
    else:
        raise ValueError("Invalid dimensionality")

    return primitive_state


# ================= Wind injection schemes =================

# here we implement all the injection schemes from
# https://arxiv.org/abs/2107.14673


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["num_ghost_cells", "num_injection_cells"])
def _wind_meo(
    wind_params: WindParams,
    primitive_state: Float[Array, "num_vars num_cells"],
    dt: Float[Array, ""],
    helper_data: HelperData,
    num_ghost_cells: int,
    num_injection_cells: int,
    gamma: Union[float, Float[Array, ""]],
) -> Float[Array, "num_vars num_cells"]:
    """Inject stellar wind into the simulation by a mass-and-energy-overwrite scheme (MEO).

    Args:
        wind_params: The wind parameters.
        primitive_state: The primitive state array.
        dt: The time step.
        helper_data: The helper data.
        num_ghost_cells: The number of ghost cells.
        num_injection_cells: The number of injection cells.
        gamma: The adiabatic index.

    Returns:
        The primitive state array with the stellar wind injected.
    """

    # set density
    density_overwrite = (
        wind_params.wind_mass_loss_rate
        / helper_data.cell_volumes[
            num_ghost_cells : num_injection_cells + num_ghost_cells
        ]
        / wind_params.wind_final_velocity
        * (
            helper_data.outer_cell_boundaries[
                num_ghost_cells : num_injection_cells + num_ghost_cells
            ]
            - helper_data.inner_cell_boundaries[
                num_ghost_cells : num_injection_cells + num_ghost_cells
            ]
        )
    )
    primitive_state = primitive_state.at[
        0, num_ghost_cells : num_injection_cells + num_ghost_cells
    ].set(density_overwrite)

    # set velocity
    primitive_state = primitive_state.at[
        1, num_ghost_cells : num_injection_cells + num_ghost_cells
    ].set(wind_params.wind_final_velocity)

    # set pressure to the floor value
    primitive_state = primitive_state.at[
        2, num_ghost_cells : num_injection_cells + num_ghost_cells
    ].set(wind_params.pressure_floor)

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit,
    static_argnames=[
        "config",
        "num_ghost_cells",
        "num_injection_cells",
        "registered_variables",
    ],
)
def _wind_mei(
    wind_params: WindParams,
    primitive_state: Float[Array, "num_vars num_cells"],
    dt: Float[Array, ""],
    config: SimulationConfig,
    helper_data: HelperData,
    num_ghost_cells: int,
    num_injection_cells: int,
    gamma: Union[float, Float[Array, ""]],
    registered_variables: RegisteredVariables,
) -> Float[Array, "num_vars num_cells"]:
    """Inject stellar wind into the simulation by a momentum-and-energy-injection scheme (MEI).

    Args:
        wind_params: The wind parameters.
        primitive_state: The primitive state array.
        dt: The time step.
        helper_data: The helper data.
        num_ghost_cells: The number of ghost cells.
        num_injection_cells: The number of injection cells.
        gamma: The adiabatic index.

    Returns:
        The primitive state array with the stellar wind injected.
    """

    conservative_state = conserved_state_from_primitive(
        primitive_state, gamma, config, registered_variables
    )

    V_inj = (
        4
        / 3
        * jnp.pi
        * helper_data.outer_cell_boundaries[num_injection_cells + num_ghost_cells] ** 3
    )

    drho = wind_params.wind_mass_loss_rate * dt / V_inj
    dmomentum = wind_params.wind_final_velocity * drho
    denergy = 0.5 * wind_params.wind_final_velocity**2 * drho

    conservative_state = conservative_state.at[
        0, num_ghost_cells : num_injection_cells + num_ghost_cells
    ].add(drho)
    conservative_state = conservative_state.at[
        1, num_ghost_cells : num_injection_cells + num_ghost_cells
    ].add(dmomentum)
    conservative_state = conservative_state.at[
        2, num_ghost_cells : num_injection_cells + num_ghost_cells
    ].add(denergy)

    primitive_state = primitive_state_from_conserved(
        conservative_state, gamma, config, registered_variables
    )

    return primitive_state


# not really ei
# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit,
    static_argnames=["num_ghost_cells", "num_injection_cells", "registered_variables"],
)
def _wind_ei(
    wind_params: WindParams,
    primitive_state: Float[Array, "num_vars num_cells"],
    dt: Float[Array, ""],
    helper_data: HelperData,
    num_ghost_cells: int,
    num_injection_cells: int,
    gamma: Union[float, Float[Array, ""]],
    registered_variables: RegisteredVariables,
) -> Float[Array, "num_vars num_cells"]:
    """Inject stellar wind into the simulation by an thermal-energy-injection scheme (EI).

    Args:
        wind_params: The wind parameters.
        primitive_state: The primitive state array.
        dt: The time step.
        helper_data: The helper data.
        num_ghost_cells: The number of ghost cells.
        num_injection_cells: The number of injection cells.
        gamma: The adiabatic index.

    Returns:
        The primitive state array with the stellar wind injected.
    """

    source_term = jnp.zeros_like(primitive_state)

    # r = helper_data.volumetric_centers
    # r = helper_data.outer_cell_boundaries
    # r_inj = r[num_injection_cells + 2]
    # V = 4/3 * jnp.pi * r_inj**3

    V = jnp.sum(
        helper_data.cell_volumes[
            num_ghost_cells : num_injection_cells + num_ghost_cells
        ]
    )

    # mass injection
    drho_dt = wind_params.wind_mass_loss_rate / V
    source_term = source_term.at[
        0, num_ghost_cells : num_injection_cells + num_ghost_cells
    ].set(drho_dt)
    updated_density = (
        primitive_state[0, num_ghost_cells : num_injection_cells + num_ghost_cells]
        + drho_dt * dt
    )

    if registered_variables.wind_density_active:
        source_term = source_term.at[
            registered_variables.wind_density_index,
            num_ghost_cells : num_injection_cells + num_ghost_cells,
        ].set(drho_dt)

    # energy injection
    dE_dt = (
        0.5 * wind_params.wind_final_velocity**2 * wind_params.wind_mass_loss_rate / V
    )

    dp_dt = pressure_from_energy(
        dE_dt,
        updated_density,
        primitive_state[1, num_ghost_cells : num_injection_cells + num_ghost_cells],
        gamma,
    )

    source_term = source_term.at[
        2, num_ghost_cells : num_injection_cells + num_ghost_cells
    ].set(dp_dt)

    primitive_state = primitive_state + source_term * dt

    return primitive_state


# not really ei
# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit,
    static_argnames=["num_ghost_cells", "num_injection_cells", "registered_variables"],
)
def dummy_multi_star_wind(
    wind_params: WindParams,
    primitive_state: STATE_TYPE,
    dt: Float[Array, ""],
    config: SimulationConfig,
    helper_data: HelperData,
    num_ghost_cells: int,
    num_injection_cells: int,
    gamma: Union[float, Float[Array, ""]],
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    star_positions = [
        jnp.array([0.2, 0.3, 0.5]),
        jnp.array([0.5, 0.7, 0.5]),
        jnp.array([0.7, 0.4, 0.5]),
        jnp.array([0.3, 0.6, 0.5]),
    ]

    for star_position in star_positions:
        r = jnp.linalg.norm(helper_data.geometric_centers - star_position, axis=-1)

        source_term = jnp.zeros_like(primitive_state)

        r_inj = num_injection_cells * config.grid_spacing
        V = 4 / 3 * jnp.pi * r_inj**3

        # for now only allow injection at the box center
        injection_mask = r <= r_inj - config.grid_spacing / 2

        # mass injection
        drho_dt = wind_params.wind_mass_loss_rate / V
        # source_term = source_term.at[registered_variables.density_index].set(jnp.where(injection_mask, drho_dt, source_term[registered_variables.density_index]))
        source_term = source_term.at[registered_variables.density_index].set(
            drho_dt * injection_mask
        )

        updated_density = primitive_state[registered_variables.density_index]
        updated_density = jnp.where(
            injection_mask > 0,
            updated_density + drho_dt * dt * injection_mask,
            updated_density,
        )

        # energy injection
        dE_dt = (
            0.5
            * wind_params.wind_final_velocity**2
            * wind_params.wind_mass_loss_rate
            / V
        )
        u = jnp.sqrt(
            primitive_state[registered_variables.velocity_index.x] ** 2
            + primitive_state[registered_variables.velocity_index.y] ** 2
            + primitive_state[registered_variables.velocity_index.z] ** 2
        )
        dp_dt = pressure_from_energy(dE_dt, updated_density, u, gamma)

        source_term = source_term.at[registered_variables.pressure_index].set(
            dp_dt * injection_mask
        )

        primitive_state = primitive_state + source_term * dt

    return primitive_state


# not really ei
# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit,
    static_argnames=["num_ghost_cells", "num_injection_cells", "registered_variables"],
)
def _wind_ei3D(
    wind_params: WindParams,
    primitive_state: STATE_TYPE,
    dt: Float[Array, ""],
    config: SimulationConfig,
    helper_data: HelperData,
    num_ghost_cells: int,
    num_injection_cells: int,
    gamma: Union[float, Float[Array, ""]],
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    """Inject stellar wind into the simulation by an thermal-energy-injection scheme (EI).

    Args:
        wind_params: The wind parameters.
        primitive_state: The primitive state array.
        dt: The time step.
        helper_data: The helper data.
        num_ghost_cells: The number of ghost cells.
        num_injection_cells: The number of injection cells.
        gamma: The adiabatic index.

    Returns:
        The primitive state array with the stellar wind injected.
    """

    source_term = jnp.zeros_like(primitive_state)

    r_inj = num_injection_cells * config.grid_spacing
    V = 4 / 3 * jnp.pi * r_inj**3

    # for now only allow injection at the box center
    injection_mask = helper_data.r <= r_inj - config.grid_spacing / 2
    # overlap_weights = (r_inj + config.grid_spacing / 2 - helper_data.r) / config.grid_spacing
    # overlap_mask = (helper_data.r > r_inj - config.grid_spacing / 2) & (helper_data.r < r_inj + config.grid_spacing / 2)
    # overlap_weights = overlap_weights * overlap_mask
    # injection_mask = injection_mask | overlap_mask
    # injection_mask = injection_mask / jnp.sum(injection_mask * config.grid_spacing**3) * V

    # mass injection
    drho_dt = wind_params.wind_mass_loss_rate / V
    # source_term = source_term.at[registered_variables.density_index].set(jnp.where(injection_mask, drho_dt, source_term[registered_variables.density_index]))
    source_term = source_term.at[registered_variables.density_index].set(
        drho_dt * injection_mask
    )

    updated_density = primitive_state[registered_variables.density_index]
    updated_density = jnp.where(
        injection_mask > 0,
        updated_density + drho_dt * dt * injection_mask,
        updated_density,
    )

    # scale down the velocity in the primitive state to conserve momentum
    # density_ratio = updated_density / primitive_state[registered_variables.density_index]
    # primitive_state = primitive_state.at[registered_variables.velocity_index.x].set(jnp.where(injection_mask, primitive_state[registered_variables.velocity_index.x] * density_ratio, primitive_state[registered_variables.velocity_index.x]))
    # primitive_state = primitive_state.at[registered_variables.velocity_index.y].set(jnp.where(injection_mask, primitive_state[registered_variables.velocity_index.y] * density_ratio, primitive_state[registered_variables.velocity_index.y]))
    # primitive_state = primitive_state.at[registered_variables.velocity_index.z].set(jnp.where(injection_mask, primitive_state[registered_variables.velocity_index.z] * density_ratio, primitive_state[registered_variables.velocity_index.z]))

    # energy injection
    dE_dt = (
        0.5 * wind_params.wind_final_velocity**2 * wind_params.wind_mass_loss_rate / V
    )
    u = jnp.sqrt(
        primitive_state[registered_variables.velocity_index.x] ** 2
        + primitive_state[registered_variables.velocity_index.y] ** 2
        + primitive_state[registered_variables.velocity_index.z] ** 2
        + 1e-20
    )
    dp_dt = pressure_from_energy(dE_dt, updated_density, u, gamma)

    source_term = source_term.at[registered_variables.pressure_index].set(
        dp_dt * injection_mask
    )

    primitive_state = primitive_state + source_term * dt

    return primitive_state

@partial(
    jax.jit,
    static_argnames=["num_injection_cells", "registered_variables", "config"],
)
def _wind_ei3D_source(
    wind_params: WindParams,
    conserved_state: STATE_TYPE,
    dt: Float[Array, ""],
    config: SimulationConfig,
    helper_data: HelperData,
    num_injection_cells: int,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:

    source_term = jnp.zeros_like(conserved_state)

    r_inj = num_injection_cells * config.grid_spacing
    r_1 = 1.3 * r_inj

    # taper down injection to avoid sharp cut-off
    # injection_mask = helper_data.r <= r_inj
    injection_mask = jnp.where(
        helper_data.r <= r_inj,
        1.0,
        jnp.where(
            (helper_data.r > r_inj) & (helper_data.r <= r_1),
            (r_1 - helper_data.r) / (r_1 - r_inj),
            0.0,
        ),
    )

    V = jnp.sum(injection_mask) * config.grid_spacing**3

    # mass injection
    # V = sum(injection_mask) * config.grid_spacing**3
    drho_dt = wind_params.wind_mass_loss_rate / V

    source_term = source_term.at[registered_variables.density_index].set(
        drho_dt * injection_mask * dt
    )

    # energy injection
    dE = (
        0.5 * wind_params.wind_final_velocity**2 * wind_params.wind_mass_loss_rate / V * dt
    )

    # we do not want to inject kinetic energy, only thermal, the kinetic energy
    # is 1/2 * rho * v^2 = 1/2 * m^2 / rho, momentum m, as we change the density,
    # and until now keep the momentum constant, we would change the kinetic energy
    # to keep the kinetic energy constant
    # 1/2 m_old^2 / rho_old = 1/2 m_new^2 / rho_new ->
    # m_new = m_old * sqrt(rho_new / rho_old) ->
    # dm = m_old * (sqrt(rho_new / rho_old) - 1)
    #    = m_old * (sqrt((rho_old + drho * dt) / rho_old) - 1)
    #    = m_old * (sqrt(1 + drho * dt / rho_old) - 1)
    momentum_source_factor = jnp.sqrt(
        1 + drho_dt * dt * injection_mask
        / (conserved_state[registered_variables.density_index])
    ) - 1.0
    # ensure we only inject momentum within r_1, while
    # the momentum source factor should be zero outside r_1 anyway
    # (sqrt(1 + 0) - 1) = 0 there might be small numerical errors
    momentum_source_factor = jnp.where(
        helper_data.r <= r_1, momentum_source_factor, 0.0
    )

    source_term = source_term.at[registered_variables.momentum_index.x].set(
        conserved_state[registered_variables.momentum_index.x] * momentum_source_factor
    )
    source_term = source_term.at[registered_variables.momentum_index.y].set(
        conserved_state[registered_variables.momentum_index.y] * momentum_source_factor
    )
    source_term = source_term.at[registered_variables.momentum_index.z].set(
        conserved_state[registered_variables.momentum_index.z] * momentum_source_factor
    )

    source_term = source_term.at[registered_variables.energy_index].set(
        dE * injection_mask
    )

    return source_term


# # @jaxtyped(typechecker=typechecker)
# @partial(jax.jit, static_argnames=['num_ghost_cells', 'num_injection_cells', 'registered_variables'])
# def _wind_ei3D_superres(wind_params: WindParams, primitive_state: STATE_TYPE, dt: Float[Array, ""], config: SimulationConfig, helper_data: HelperData, num_ghost_cells: int, num_injection_cells: int, gamma: Union[float, Float[Array, ""]], registered_variables: RegisteredVariables) -> STATE_TYPE:
#     """Inject stellar wind into the simulation by an thermal-energy-injection scheme (EI).

#     Args:
#         wind_params: The wind parameters.
#         primitive_state: The primitive state array.
#         dt: The time step.
#         helper_data: The helper data.
#         num_ghost_cells: The number of ghost cells.
#         num_injection_cells: The number of injection cells.
#         gamma: The adiabatic index.

#     Returns:
#         The primitive state array with the stellar wind injected.
#     """

#     source_term = jnp.zeros_like(primitive_state)
#     r_inj = num_injection_cells * config.grid_spacing

#     total_mass_change = wind_params.wind_mass_loss_rate * dt
#     total_energy_change = 0.5 * wind_params.wind_final_velocity**2 * total_mass_change

#     superres_factor = 8
#     superres_grid_size = superres_factor * num_injection_cells * 2
#     superres_grid_spacing = config.grid_spacing / superres_factor

#     half_width = superres_grid_size * superres_grid_spacing / 2

#     x = jnp.linspace(-half_width, half_width, superres_grid_size)
#     y = jnp.linspace(-half_width, half_width, superres_grid_size)
#     z = jnp.linspace(-half_width, half_width, superres_grid_size)
#     X, Y, Z = jnp.meshgrid(x, y, z, indexing='ij')
#     R = jnp.sqrt(X**2 + Y**2 + Z**2)
#     superres_injection_weights = R <= r_inj
#     superres_injection_weights = superres_injection_weights / jnp.sum(superres_injection_weights)


#     # sum pool down the mask to get to a mask of size (num_injection_cells * 2)^3
#     superres_injection_weights = superres_injection_weights.reshape((num_injection_cells * 2, superres_factor,
#                                    num_injection_cells * 2, superres_factor,
#                                    num_injection_cells * 2, superres_factor)).sum(axis=(1, 3, 5))


#     injection_weights = jnp.zeros_like(primitive_state[0])
#     half_index = primitive_state[0].shape[0] // 2
#     injection_weights = injection_weights.at[half_index - num_injection_cells:half_index + num_injection_cells, half_index - num_injection_cells:half_index + num_injection_cells, half_index - num_injection_cells:half_index + num_injection_cells].set(superres_injection_weights)

#     source_term = source_term.at[registered_variables.density_index].set(total_mass_change * injection_weights / (config.grid_spacing**3))
#     gamma = 4/3
#     source_term = source_term.at[registered_variables.pressure_index].set(total_energy_change * (gamma - 1) * injection_weights / (config.grid_spacing**3))

#     primitive_state = primitive_state + source_term

#     return primitive_state

# ======================================================



# general
from functools import partial
import jax.numpy as jnp
import jax

import equinox as eqx

# typing
from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker
from typing import Tuple, Union

from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import STATE_TYPE, SimulationConfig
from astronomix.option_classes.simulation_params import SimulationParams


class ForceNet(eqx.Module):
    mlp: eqx.nn.MLP

    def __init__(self, key):
        self.mlp = eqx.nn.MLP(
            in_size=3,  # <- now takes (x, y, t)
            out_size=2,  # returns (Fx, Fy)
            width_size=128,
            depth=4,
            key=key,
        )

    def __call__(self, xyt):
        return self.mlp(xyt)


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _neural_net_force(
    primitive_state: STATE_TYPE,
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    params: SimulationParams,
    helper_data: HelperData,
    time_step: Float[Array, ""],
    current_time: Float[Array, ""],  # <- new argument
):
    neural_net_params = params.neural_net_force_params.network_params
    neural_net_static = config.neural_net_force_config.network_static
    model = eqx.combine(neural_net_params, neural_net_static)

    N = config.num_cells
    positions = helper_data.geometric_centers
    positions_flat = positions.reshape(-1, 2)  # (N*N, 2)

    # Add time dimension: broadcast time to (N*N, 1) and concat
    time_broadcasted = jnp.full((positions_flat.shape[0], 1), current_time)
    positions_with_time = jnp.concatenate(
        [positions_flat, time_broadcasted], axis=1
    )  # (N*N, 3)

    # Apply model to (x, y, t)
    forces_flat = jax.vmap(model)(positions_with_time)  # shape (N*N, 2)

    # Reshape back to (2, N, N)
    forces = forces_flat.reshape(N, N, 2).transpose(2, 0, 1)

    nc = config.num_ghost_cells
    primitive_state = primitive_state.at[
        registered_variables.velocity_index.x, nc:-nc, nc:-nc
    ].add(forces[0] * time_step)
    primitive_state = primitive_state.at[
        registered_variables.velocity_index.y, nc:-nc, nc:-nc
    ].add(forces[1] * time_step)

    return primitive_state



from astronomix.data_classes.simulation_snapshot_data import SnapshotData
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.data_classes.simulation_helper_data import get_helper_data


# general imports
import jax.numpy as jnp
import jax
from functools import partial


# type checking imports
from jaxtyping import Array, Float
from beartype import beartype as typechecker
from typing import Union

# general astronomix imports
from astronomix._finite_difference._fluid_equations._equations import conserved_state_from_primitive_mhd, primitive_state_from_conserved_mhd
from astronomix._finite_difference._time_integrators._ssprk import _ssprk4_with_ct
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import (
    STATE_TYPE,
    SimulationConfig,
)

from astronomix.option_classes.simulation_params import SimulationParams

@partial(jax.jit, static_argnames=["config", "registered_variables"], donate_argnames=["primitive_state"])
def _evolve_state_fd(
    primitive_state: STATE_TYPE,
    dt: Float[Array, ""],
    gamma: Union[float, Float[Array, ""]],
    gravitational_constant: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    params: SimulationParams,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    
    # NOTE: here we assume the magnetic field at interfaces
    # is stored in the last three indices of the state array
    conserved_state = conserved_state_from_primitive_mhd(
        primitive_state[:-3], gamma, registered_variables
    )

    # extract interface magnetic fields
    bxb = primitive_state[registered_variables.interface_magnetic_field_index.x]
    byb = primitive_state[registered_variables.interface_magnetic_field_index.y]
    bzb = primitive_state[registered_variables.interface_magnetic_field_index.z]

    # update conserved state and interface magnetic fields
    conserved_state, bxb, byb, bzb = _ssprk4_with_ct(
        conserved_state,
        bxb,
        byb,
        bzb,
        gamma,
        config.grid_spacing,
        dt,
        params,
        helper_data,
        config,
        registered_variables,
    )

    # back to primitive state
    primitive_state = primitive_state_from_conserved_mhd(
        conserved_state, params.minimum_density, params.minimum_pressure, gamma, config, registered_variables
    )

    # append updated interface magnetic fields
    # NOTE: same assumption as above
    primitive_state = jnp.concatenate(
        [primitive_state, bxb[None, :], byb[None, :], bzb[None, :]], axis=0
    )

    return primitive_state


import jax
import jax.numpy as jnp


from functools import partial


@partial(jax.jit, static_argnames=["axis"])
def finite_difference_int6(f_int, axis):
    """
    High-order FD derivative, assumed f defined at interfaces,
    and the i-th index corresponds to the i+1/2 interface.

    The finite difference formula is:
        df/dx at i = c1 * (f_{i+1/2} - f_{i-1/2})
                     + c2 * (f_{i+3/2} - f_{i-3/2})
                     + c3 * (f_{i+5/2} - f_{i-5/2})

    Note that the i+1/2 interface corresponds
    to index i in the array.

    6th order: c1 = 75/64, c2 = -25/384, c3 = 3/640
    4th order: c1 = 9/8, c2 = -1/24, c3 = 0
    """
    c1, c2, c3 = 75.0 / 64.0, -25.0 / 384.0, 3.0 / 640.0
    return (
        c1 * (f_int - jnp.roll(f_int, 1, axis=axis))
        + c2 * (jnp.roll(f_int, -1, axis=axis) - jnp.roll(f_int, 2, axis=axis))
        + c3 * (jnp.roll(f_int, -2, axis=axis) - jnp.roll(f_int, 3, axis=axis))
    )

@jax.jit
def _interface_field_divergence(bx_int, by_int, bz_int, grid_spacing):
    """
    Compute the divergence of the magnetic field
    defined at interfaces using finite difference.

    Args:
        bx_int: Magnetic field in x-direction at x interfaces.
        by_int: Magnetic field in y-direction at y interfaces.
        bz_int: Magnetic field in z-direction at z interfaces.
        grid_spacing: Grid spacing (assumed uniform in all directions).

    Returns:
        Divergence of the magnetic field at cell centers.
    """
    d_bx_dx = finite_difference_int6(bx_int, axis=0) / grid_spacing
    d_by_dy = finite_difference_int6(by_int, axis=1) / grid_spacing
    d_bz_dz = finite_difference_int6(bz_int, axis=2) / grid_spacing

    div_b = d_bx_dx + d_by_dy + d_bz_dz
    return div_b


# general imports
from functools import partial
import jax
import jax.numpy as jnp

# typechecking
from typing import Union
from beartype import beartype as typechecker
from jaxtyping import Array, Float, jaxtyped

# general astronomix imports
from astronomix.option_classes.simulation_config import (
    STATE_TYPE,
    STATE_TYPE_ALTERED,
    SimulationConfig,
)
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.variable_registry.registered_variables import RegisteredVariables

# limited gradients
from astronomix._finite_volume._state_evolution.limited_gradients import _calculate_limited_gradients


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _pressure_nozzling_source(
    primitive_state: STATE_TYPE,
    config: SimulationConfig,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE_ALTERED:
    """Pressure nozzling source term as of the geometry of the domain.

    Args:
        primitive_state: The primitive state array.
        config: The simulation configuration.
        helper_data: The helper data.
        registered_variables: The registered variables.

    Returns:
        The pressure nozzling source
    """

    # get the pressure
    p = primitive_state[registered_variables.pressure_index]

    # calculate the limited pressure gradients
    dp_dr = _calculate_limited_gradients(primitive_state, config, helper_data, axis=1)[
        registered_variables.pressure_index
    ]

    # calculate the pressure nozzling term, following
    # eq. 14 in Crittendend and Balachandar, 2018
    # https://doi.org/10.1007/s00193-017-0784-y
    pressure_nozzling = (
        helper_data.geometric_centers ** (config.geometry - 1) * p
        + (
            helper_data.r_hat_alpha
            - helper_data.volumetric_centers
            * helper_data.geometric_centers ** (config.geometry - 1)
        )
        * dp_dr
    )
    nozzling = jnp.zeros((registered_variables.num_vars, p.shape[0]))
    nozzling = nozzling.at[registered_variables.velocity_index].set(
        config.geometry * pressure_nozzling
    )

    return nozzling



# TOWNSEND SCHEME DOES NOT WORK CURRENTLY,
# ONLY THE SIMPLE EXPLICIT COOLING IS USED

# For proper cooling something like Grackle
# might be used. For now we are interested in
# the most simple cooling model.

# dE/dt + ... = Phi(T, rho)
# Phi(T, rho) = n_H * Gamma(T) - n_H ^ 2 * Lambda(T)

# for a simple cooling term (Lambda) see 5.3 in
# https://arxiv.org/pdf/2111.03399

# also see
# https://academic.oup.com/mnras/article/502/3/3179/6081066

# and
# https://iopscience.iop.org/article/10.1088/0067-0049/181/2/391

from typing import Tuple

from functools import partial

import jax

import equinox as eqx

from astronomix._physics_modules._cooling.cooling_options import (
    COOLING_CURVE_TYPE,
    EXPLICIT_COOLING,
    IMPLICIT_COOLING,
    NEURAL_NET_COOLING,
    NEURAL_NET_COOLING_WITH_DENSITY,
    PIECEWISE_POWER_LAW,
    SIMPLE_POWER_LAW,
    CoolingConfig,
    CoolingCurveConfig,
    CoolingParams,
)
from astronomix._finite_volume._state_evolution.limited_gradients import _calculate_limited_gradients
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import FIELD_TYPE, STATE_TYPE, SimulationConfig

import jax.numpy as jnp

from astronomix.option_classes.simulation_params import SimulationParams

# NOTE the rescaled units
# \tilde{T} = T * k_B / u
# \tilde{\Lambda} = \lambda / u^2


def get_effective_molecular_weights(
    hydrogen_mass_fraction: float,  # X
    metal_mass_fraction: float,  # Z
) -> Tuple[float, float, float]:
    """
    Calculate the mean molecular weight (mu)
    and the effective molecular weights for
    electrons (mu_e), hydrogen (mu_H)
    """

    # mean molecular weight
    mu = 1.0 / (
        2 * hydrogen_mass_fraction
        + 3 * (1 - hydrogen_mass_fraction - metal_mass_fraction) / 4
        + metal_mass_fraction / 2
    )

    # effective molecular weight for electrons
    mu_e = 2 * 1.0 / (1 + hydrogen_mass_fraction)

    # effective molecular weight for hydrogen
    mu_H = 1.0 / hydrogen_mass_fraction

    return mu, mu_e, mu_H


def get_particle_number_density(
    density: FIELD_TYPE, mean_molecular_weight: float
) -> FIELD_TYPE:
    return density / mean_molecular_weight


def get_pressure_from_temperature(
    density: FIELD_TYPE,
    temperature: FIELD_TYPE,
    hydrogen_mass_fraction: float,
    metal_mass_fraction: float,
) -> FIELD_TYPE:
    """
    P = n * \tilde{T}
    """

    # calculate the effective molecular weights
    mu, _, _ = get_effective_molecular_weights(
        hydrogen_mass_fraction,
        metal_mass_fraction,
    )

    # calculate the particle number density
    n = get_particle_number_density(density, mu)

    # calculate the pressure
    return n * temperature


def get_temperature_from_pressure(
    density: FIELD_TYPE,
    pressure: FIELD_TYPE,
    hydrogen_mass_fraction: float,
    metal_mass_fraction: float,
) -> FIELD_TYPE:
    """
    \tilde{T} = P / \tilde{n}
    """

    # calculate the effective molecular weights
    mu, _, _ = get_effective_molecular_weights(
        hydrogen_mass_fraction, metal_mass_fraction
    )

    # calculate the particle number density
    n = get_particle_number_density(density, mu)

    # calculate the temperature
    return pressure / n  # so the density must never be zero


# \Lambda(T)
def cooling_rate_power_law(
    temperature: FIELD_TYPE,
    reference_temperature: float,
    factor: float,
    exponent: float,
):
    return factor * (temperature / reference_temperature) ** exponent


# t_cool
@partial(jax.jit, static_argnames=("cooling_curve_config",))
def cooling_time(
    density: FIELD_TYPE,
    temperature: FIELD_TYPE,
    hydrogen_mass_fraction: float,
    metal_mass_fraction: float,
    gamma: float,
    cooling_curve_config: CoolingCurveConfig,
    cooling_curve_params: COOLING_CURVE_TYPE,
) -> FIELD_TYPE:
    """
    t_cool = (k * mu_e * mu_H * T) / ((gamma - 1) * rho * mu * Lambda(T))
    """

    # calculate the effective molecular weights
    mu, mu_e, mu_H = get_effective_molecular_weights(
        hydrogen_mass_fraction,
        metal_mass_fraction,
    )

    # calculate the cooling rate
    cooling_rate = _cooling_rate(
        temperature,
        density,
        cooling_curve_config,
        cooling_curve_params,
    )

    # calculate the cooling time
    return (mu_e * mu_H * temperature) / ((gamma - 1) * density * mu * cooling_rate)


# Y(T)
def power_law_temporal_evolution_function(
    temperature: FIELD_TYPE,  # T
    reference_temperature: float,  # T_ref
    exponent: float,  # alpha
) -> FIELD_TYPE:
    """
    1/(1 - alpha) * (1 - (T/T_ref)^(1-alpha)) for alpha != 1
    -log(T/T_ref) for alpha = 1
    """
    return jax.lax.cond(
        exponent != 1,
        lambda: (1 / (1 - exponent))
        * (1 - (temperature / reference_temperature) ** (1 - exponent)),
        lambda: -jnp.log(temperature / reference_temperature),
    )


# Y^-1(Y)
def power_law_temporal_evolution_function_inverse(
    temporal_evolution_function: FIELD_TYPE,  # Y
    reference_temperature: float,  # T_ref
    exponent: float,  # alpha
) -> FIELD_TYPE:
    """
    T_ref * (1 - (1 - alpha) * Y)^(1/(1-alpha)) for alpha != 1
    T_ref * exp(-Y) for alpha = 1
    """
    return jax.lax.cond(
        exponent != 1,
        lambda: reference_temperature
        * (1 - (1 - exponent) * temporal_evolution_function) ** (1 / (1 - exponent)),
        lambda: reference_temperature * jnp.exp(-temporal_evolution_function),
    )


# piecewise power law
@partial(
    jnp.vectorize,
    excluded=(1, 2, 3),  # don’t vectorize over tables
    signature="()->()",  # scalar in, scalar out
)
def _evaluate_piecewise_power_law(
    T_in,
    T_table,
    Lambda_table,
    alpha_table,
):
    def eval_in_range(T_in):
        k = jnp.searchsorted(T_table, T_in) - 1

        # clip k to be in the valid range
        k = jnp.clip(k, 0, len(T_table) - 2)

        alpha_k = alpha_table[k]
        T_k = T_table[k]
        Lambda_k = Lambda_table[k]
        return Lambda_k * (T_in / T_k) ** alpha_k

    return jax.lax.cond(
        # check if T_in is in the table range
        (T_in >= T_table[0]) & (T_in <= T_table[-1]),
        eval_in_range,
        lambda _: 0.0,  # return 0 if out of range
        T_in,
    )


@partial(
    jnp.vectorize,
    excluded=(1, 2, 3, 4),  # don’t vectorize over tables
    signature="()->()",  # scalar in, scalar out
)
def _piecewise_power_law_temporal_evolution_function(
    T_in, T_table, Lambda_table, alpha_table, Y_table
):
    def eval_in_range(T_in):
        k = jnp.searchsorted(T_table, T_in) - 1

        # clip k to be in the valid range
        k = jnp.clip(k, 0, len(T_table) - 2)

        alpha_k = alpha_table[k]
        T_k = T_table[k]
        Lambda_k = Lambda_table[k]
        Y_k = Y_table[k]
        return Y_k + jax.lax.cond(
            alpha_k != 1.0,
            lambda: 1
            / (1 - alpha_k)
            * Lambda_table[-1]
            / Lambda_k
            * T_k
            / T_table[-1]
            * (1 - (T_k / T_in) ** (alpha_k - 1)),
            lambda: Lambda_table[-1]
            / Lambda_k
            * T_k
            / T_table[-1]
            * jnp.log(T_k / T_in),
        )

    return jax.lax.cond(
        # check if T_in is in the table range
        (T_in >= T_table[0]) & (T_in <= T_table[-1]),
        eval_in_range,
        lambda _: 0.0,  # return 0 if out of range
        T_in,
    )


@partial(
    jnp.vectorize,
    excluded=(1, 2, 3, 4),  # don’t vectorize over tables
    signature="()->()",  # scalar in, scalar out
)
def _piecewise_power_law_temporal_evolution_function_inverse(
    Y_in, T_table, Lambda_table, alpha_table, Y_table
):
    def eval_in_range(Y_in):
        # k such that Y_k >= Y >= Y_{k+1}
        k = jnp.searchsorted(-Y_table, -Y_in) - 1

        # clip k to be in the valid range
        k = jnp.clip(k, 0, len(Y_table) - 2)

        alpha_k = alpha_table[k]
        T_k = T_table[k]
        Lambda_k = Lambda_table[k]
        Y_k = Y_table[k]
        return jax.lax.cond(
            alpha_k != 1.0,
            lambda: T_k
            * (
                1
                - (1 - alpha_k)
                * (Y_in - Y_k)
                * Lambda_k
                / Lambda_table[-1]
                * T_table[-1]
                / T_k
            )
            ** (1 / (1 - alpha_k)),
            lambda: T_k
            * jnp.exp(-(Y_in - Y_k) * Lambda_k / Lambda_table[-1] * T_table[-1] / T_k),
        )

    return jax.lax.cond(
        # check if Y_in is in the table range,
        # Y_table is monotonically decreasing
        (Y_in >= Y_table[-1]) & (Y_in <= Y_table[0]),
        eval_in_range,
        lambda _: jnp.where(Y_in < Y_table[-1], T_table[-1], T_table[0]),
        Y_in,
    )


@partial(jax.jit, static_argnames=("cooling_curve_config",))
def _cooling_rate(
    temperature: FIELD_TYPE,
    density: FIELD_TYPE,
    cooling_curve_config: CoolingCurveConfig,
    cooling_curve_params: COOLING_CURVE_TYPE,
) -> FIELD_TYPE:
    if cooling_curve_config.cooling_curve_type == SIMPLE_POWER_LAW:
        return cooling_rate_power_law(
            temperature,
            cooling_curve_params.reference_temperature,
            cooling_curve_params.factor,
            cooling_curve_params.exponent,
        )
    elif cooling_curve_config.cooling_curve_type == PIECEWISE_POWER_LAW:
        return _evaluate_piecewise_power_law(
            temperature,
            10**cooling_curve_params.log10_T_table,
            10**cooling_curve_params.log10_Lambda_table,
            cooling_curve_params.alpha_table,
        )
    elif cooling_curve_config.cooling_curve_type == NEURAL_NET_COOLING:
        neural_net_params = cooling_curve_params.network_params
        neural_net_static = cooling_curve_config.cooling_net_config.network_static
        model = jax.vmap(eqx.combine(neural_net_params, neural_net_static))

        # for now we train the network in the specific code units,
        # so no appropriate rescaling here, to be changed later
        return 10 ** model(jnp.log10(temperature).reshape(-1, 1)).flatten()
    elif cooling_curve_config.cooling_curve_type == NEURAL_NET_COOLING_WITH_DENSITY:
        neural_net_params = cooling_curve_params.network_params
        neural_net_static = cooling_curve_config.cooling_net_config.network_static
        model = jax.vmap(eqx.combine(neural_net_params, neural_net_static))

        # for now we train the network in the specific code units,
        # so no appropriate rescaling here, to be changed later
        input_data = jnp.stack([jnp.log10(temperature), jnp.log10(density)], axis=-1)
        return 10 ** model(input_data).flatten()

    else:
        raise ValueError(
            f"Unknown cooling curve type: {cooling_curve_config.cooling_curve_type}"
        )


@partial(jax.jit, static_argnames=("cooling_curve_config",))
def _temporal_evolution_function(
    temperature: FIELD_TYPE,
    cooling_curve_config: CoolingCurveConfig,
    cooling_curve_params: COOLING_CURVE_TYPE,
) -> FIELD_TYPE:
    if cooling_curve_config.cooling_curve_type == SIMPLE_POWER_LAW:
        return power_law_temporal_evolution_function(
            temperature,
            cooling_curve_params.reference_temperature,
            cooling_curve_params.exponent,
        )
    elif cooling_curve_config.cooling_curve_type == PIECEWISE_POWER_LAW:
        return _piecewise_power_law_temporal_evolution_function(
            temperature,
            10**cooling_curve_params.log10_T_table,
            10**cooling_curve_params.log10_Lambda_table,
            cooling_curve_params.alpha_table,
            cooling_curve_params.Y_table,
        )
    else:
        raise ValueError(
            f"Unknown cooling curve type: {cooling_curve_config.cooling_curve_type}"
        )


@partial(jax.jit, static_argnames=("cooling_curve_config",))
def _temporal_evolution_function_inverse(
    temporal_evolution_function: FIELD_TYPE,
    cooling_curve_config: CoolingCurveConfig,
    cooling_curve_params: COOLING_CURVE_TYPE,
) -> FIELD_TYPE:
    if cooling_curve_config.cooling_curve_type == SIMPLE_POWER_LAW:
        return power_law_temporal_evolution_function_inverse(
            temporal_evolution_function,
            cooling_curve_params.reference_temperature,
            cooling_curve_params.exponent,
        )
    elif cooling_curve_config.cooling_curve_type == PIECEWISE_POWER_LAW:
        return _piecewise_power_law_temporal_evolution_function_inverse(
            temporal_evolution_function,
            10**cooling_curve_params.log10_T_table,
            10**cooling_curve_params.log10_Lambda_table,
            cooling_curve_params.alpha_table,
            cooling_curve_params.Y_table,
        )
    else:
        raise ValueError(
            f"Unknown cooling curve type: {cooling_curve_config.cooling_curve_type}"
        )


@partial(jax.jit, static_argnames=("cooling_curve_config",))
def update_temperature(
    density: FIELD_TYPE,
    temperature: FIELD_TYPE,
    time_step: float,
    hydrogen_mass_fraction: float,
    metal_mass_fraction: float,
    gamma: float,
    cooling_curve_config: CoolingCurveConfig,
    cooling_curve_params: COOLING_CURVE_TYPE,
) -> FIELD_TYPE:
    """
    T_new = Y^-1[Y(T) + T / T_ref * \Lambda(T_ref) / \Lambda(T) * delta_t / t_cool]
    """

    reference_temperature = cooling_curve_params.reference_temperature

    # calculate the cooling time
    # not numerically stable, divisiion
    # by the cooling
    # t_cool = cooling_time(
    #     density,
    #     temperature,
    #     hydrogen_mass_fraction,
    #     metal_mass_fraction,
    #     gamma,
    #     cooling_curve_config,
    #     cooling_curve_params
    # )

    # calculate the cooling rate
    cooling_rate = _cooling_rate(
        temperature, density, cooling_curve_config, cooling_curve_params
    )

    cooling_rate_reference = _cooling_rate(
        jnp.array([reference_temperature]),
        jnp.array([density]),
        cooling_curve_config,
        cooling_curve_params,
    )

    # calculate the temporal evolution function
    temporal_evolution_function = _temporal_evolution_function(
        temperature, cooling_curve_config, cooling_curve_params
    )

    # calculate the new temperature
    # new_temperature = _temporal_evolution_function_inverse(
    #     temporal_evolution_function + (temperature / reference_temperature) * (cooling_rate_reference / cooling_rate) * time_step / t_cool,
    #     cooling_curve_config,
    #     cooling_curve_params
    # )

    mu, mu_e, mu_H = get_effective_molecular_weights(
        hydrogen_mass_fraction,
        metal_mass_fraction,
    )

    new_temperature = _temporal_evolution_function_inverse(
        temporal_evolution_function
        + cooling_rate_reference
        / reference_temperature
        * ((gamma - 1) * density * mu)
        / (mu_e * mu_H)
        * time_step,
        cooling_curve_config,
        cooling_curve_params,
    )

    return new_temperature


@partial(jax.jit, static_argnames=("cooling_curve_config",))
def dtemperature_dt(
    density: FIELD_TYPE,
    temperature: FIELD_TYPE,
    hydrogen_mass_fraction: float,
    metal_mass_fraction: float,
    gamma: float,
    cooling_curve_config: CoolingCurveConfig,
    cooling_curve_params: COOLING_CURVE_TYPE,
) -> FIELD_TYPE:
    """
    T_new = T - (gamma - 1) * rho * \mu / (mu_e * mu_H * k) * Lambda(T) * delta_t
    (units absorbed in Lambda)
    """

    # calculate the cooling rate
    cooling_rate = _cooling_rate(
        temperature, density, cooling_curve_config, cooling_curve_params
    )

    mu, mu_e, mu_H = get_effective_molecular_weights(
        hydrogen_mass_fraction,
        metal_mass_fraction,
    )

    return -(cooling_rate * (gamma - 1) * density * mu) / (mu_e * mu_H)


@partial(jax.jit, static_argnames=("cooling_curve_config",))
def update_temperature_explicit(
    density: FIELD_TYPE,
    temperature: FIELD_TYPE,
    time_step: float,
    hydrogen_mass_fraction: float,
    metal_mass_fraction: float,
    gamma: float,
    cooling_curve_config: CoolingCurveConfig,
    cooling_curve_params: COOLING_CURVE_TYPE,
) -> FIELD_TYPE:
    """
    T_new = T - (gamma - 1) * rho * \mu / (mu_e * mu_H * k) * Lambda(T) * delta_t
    (units absorbed in Lambda)
    """

    # t_cool = cooling_time(
    #     density,
    #     temperature,
    #     hydrogen_mass_fraction,
    #     metal_mass_fraction,
    #     gamma,
    #     cooling_curve_config,
    #     cooling_curve_params
    # )

    # dt_cool_min = jnp.minimum(time_step, 0.1 * t_cool)
    # num_steps_min = jnp.ceil(time_step / dt_cool_min)
    # dt_cool = time_step / num_steps_min

    # def T_update(T):
    #     # RK2 step
    #     k1 = dtemperature_dt(
    #         density,
    #         T,
    #         hydrogen_mass_fraction,
    #         metal_mass_fraction,
    #         gamma,
    #         cooling_curve_config,
    #         cooling_curve_params
    #     )
    #     k2 = dtemperature_dt(
    #         density,
    #         T + 0.5 * dt_cool * k1,
    #         hydrogen_mass_fraction,
    #         metal_mass_fraction,
    #         gamma,
    #         cooling_curve_config,
    #         cooling_curve_params
    #     )
    #     return T + dt_cool * k2

    # new_temperature = jax.lax.fori_loop(
    #     0, jnp.int32(num_steps_min[0]),
    #     lambda i, T: T_update(T),
    #     temperature
    # )
    # return new_temperature

    return (
        temperature
        + dtemperature_dt(
            density,
            temperature,
            hydrogen_mass_fraction,
            metal_mass_fraction,
            gamma,
            cooling_curve_config,
            cooling_curve_params,
        )
        * time_step
    )

@partial(jax.jit, static_argnames=("cooling_curve_config",))
def update_temperature_implicit(
    density: FIELD_TYPE,
    temperature: FIELD_TYPE,
    time_step: float,
    hydrogen_mass_fraction: float,
    metal_mass_fraction: float,
    gamma: float,
    cooling_curve_config: CoolingCurveConfig,
    cooling_curve_params: COOLING_CURVE_TYPE,
) -> FIELD_TYPE:

    def implicit_eq(T_new):
        return (temperature
        + dtemperature_dt(
            density,
            T_new,
            hydrogen_mass_fraction,
            metal_mass_fraction,
            gamma,
            cooling_curve_config,
            cooling_curve_params,
        ) * time_step)

    # use a simple fixed point iteration
    # - maybe do newton or bisection method later
    max_iter = 50
    tol = 1e-6

    def cond_fun(state):
        i, T_old = state
        T_candidate = implicit_eq(T_old)
        diff = jnp.max(jnp.abs(T_candidate - T_old))
        return (i < max_iter) & (diff > tol)

    def body_fun(state):
        i, T_old = state
        T_new = implicit_eq(T_old)
        return (i + 1, T_new)

    state = (0, temperature)
    _, T_final = jax.lax.while_loop(cond_fun, body_fun, state)
    return T_final


@partial(jax.jit, static_argnames=("cooling_config", "registered_variables"))
def update_pressure_by_cooling(
    primitive_state: STATE_TYPE,
    registered_variables: RegisteredVariables,
    cooling_config: CoolingConfig,
    simulation_params: SimulationParams,
    time_step: float,
) -> STATE_TYPE:
    
    cooling_curve_config = cooling_config.cooling_curve_config

    # get the parameters
    cooling_params = simulation_params.cooling_params
    hydrogen_mass_fraction = cooling_params.hydrogen_mass_fraction
    metal_mass_fraction = cooling_params.metal_mass_fraction
    gamma = simulation_params.gamma

    # get the density and pressure
    density = primitive_state[registered_variables.density_index]
    pressure = primitive_state[registered_variables.pressure_index]

    # get the temperature
    temperature = get_temperature_from_pressure(
        density,
        pressure,
        hydrogen_mass_fraction,
        metal_mass_fraction,
    )

    if cooling_config.cooling_method == IMPLICIT_COOLING:
        new_temperature = update_temperature_implicit(
            density,
            temperature,
            time_step,
            hydrogen_mass_fraction,
            metal_mass_fraction,
            gamma,
            cooling_curve_config,
            cooling_params.cooling_curve_params,
        )
    elif cooling_config.cooling_method == EXPLICIT_COOLING:
        new_temperature = update_temperature_explicit(
            density,
            temperature,
            time_step,
            hydrogen_mass_fraction,
            metal_mass_fraction,
            gamma,
            cooling_curve_config,
            cooling_params.cooling_curve_params,
        )

    new_temperature = jnp.where(
        (new_temperature > cooling_params.floor_temperature),
        new_temperature,
        temperature,
    )

    # new_temperature = temperature

    # update the pressure
    new_pressure = get_pressure_from_temperature(
        density,
        new_temperature,
        hydrogen_mass_fraction,
        metal_mass_fraction,
    )

    # set the new pressure
    primitive_state = primitive_state.at[registered_variables.pressure_index].set(
        new_pressure
    )

    # return the updated primitive state
    return primitive_state

@partial(jax.jit, static_argnames = ("config", "registered_variables"))
def first_order_pressure_update(
    primitive_state: STATE_TYPE,
    registered_variables: RegisteredVariables,
    config: SimulationConfig,
    helper_data: HelperData,
    simulation_params: SimulationParams,
    time_step: float,
) -> STATE_TYPE:
    # dP/dt_cooling = -(gamma - 1) n_e n_H Lambda(T)
    # with n_e = rho / mu_e, n_H = rho / mu_H

    cooling_curve_config = config.cooling_config.cooling_curve_config

    # get the parameters
    cooling_params = simulation_params.cooling_params
    hydrogen_mass_fraction = cooling_params.hydrogen_mass_fraction
    metal_mass_fraction = cooling_params.metal_mass_fraction
    gamma = simulation_params.gamma

    # get the density and pressure
    density = primitive_state[registered_variables.density_index]
    pressure = primitive_state[registered_variables.pressure_index]

    # get the temperature
    temperature = get_temperature_from_pressure(
        density,
        pressure,
        hydrogen_mass_fraction,
        metal_mass_fraction,
    )

    # get the molecular weights
    mu, mu_e, mu_H = get_effective_molecular_weights(
        hydrogen_mass_fraction,
        metal_mass_fraction,
    )

    # get limited gradients, support only 1d for now
    # make cartesion approximation for now
    # only along x axis for now
    limited_gradients = _calculate_limited_gradients(primitive_state, config, helper_data, axis = 1)

    # density gradient
    density_gradient = limited_gradients[registered_variables.density_index]

    # pressure gradient
    pressure_gradient = limited_gradients[registered_variables.pressure_index]

    # temperature gradient
    # n = density / mu, T = P / n = P * mu / density
    temperature_gradient = (pressure_gradient * density - pressure * density_gradient) * mu / density**2

    # finite difference approximation to the cooling rate gradient
    cooling_rate_gradient = 1/config.grid_spacing * (
        _cooling_rate(
            temperature + 0.5 * config.grid_spacing * temperature_gradient,
            density + 0.5 * config.grid_spacing * density_gradient,
            cooling_curve_config,
            cooling_params.cooling_curve_params
        ) - _cooling_rate(
            temperature - 0.5 * config.grid_spacing * temperature_gradient,
            density - 0.5 * config.grid_spacing * density_gradient,
            cooling_curve_config,
            cooling_params.cooling_curve_params
        )
    )

    # get the cooling rate
    cooling_rate = _cooling_rate(
        temperature,
        density,
        cooling_curve_config,
        cooling_params.cooling_curve_params
    )

    # update the pressure, 0th order currently
    new_pressure = (
        pressure - (gamma - 1) / (mu_e * mu_H) * time_step * (
            density**2 * cooling_rate # 0th order
            + 1/12 * config.grid_spacing ** 2 * cooling_rate * density_gradient ** 2
            + 1/6 * config.grid_spacing ** 2 * density * cooling_rate_gradient * density_gradient
        )
    )

    # jax.debug.print("mean correction = {corr}, 0th-order = {zeroth}", corr = jnp.mean(
    #     (+ 1/12 * config.grid_spacing ** 2 * cooling_rate * density_gradient ** 2
    #     + 1/6 * config.grid_spacing ** 2 * density * cooling_rate_gradient * density_gradient)
    # ), zeroth = jnp.mean(
    #     density**2 * cooling_rate
    # ))

    # calculate the new temperature
    new_temperature = get_temperature_from_pressure(
        density,
        new_pressure,
        hydrogen_mass_fraction,
        metal_mass_fraction,
    )

    # apply the temperature floor
    new_temperature = jnp.where(
        (new_temperature > cooling_params.floor_temperature),
        new_temperature,
        temperature
    )

    # recalculate the pressure
    new_pressure = get_pressure_from_temperature(
        density,
        new_temperature,
        hydrogen_mass_fraction,
        metal_mass_fraction,
    )

    # set the new pressure
    primitive_state = primitive_state.at[registered_variables.pressure_index].set(new_pressure)

    # return the updated primitive state
    return primitive_state



from functools import partial
import jax.numpy as jnp
from jaxtyping import Array, Float, jaxtyped
import jax
from equinox.internal._loop.checkpointed import checkpointed_while_loop

from astronomix._geometry.boundaries import _boundary_handler
from astronomix._finite_volume._magnetic_update._vector_maths import (
    cross,
    curl2D,
    curl3D,
    divergence2D,
)
from astronomix._fluid_equations._equations import (
    pressure_from_energy,
    total_energy_from_primitives,
)
from astronomix.variable_registry.registered_variables import RegisteredVariables

# runtime debugging
from jax.experimental import checkify

from astronomix.option_classes.simulation_config import (
    BACKWARDS,
    DOUBLE_PRECISION,
    FORWARDS,
    GHOST_CELLS,
    IMPLICIT_EULER,
    IMPLICIT_MIDPOINT,
    MAGNETIC_FIELD_ONLY,
    SINGLE_PRECISION,
    VELOCITY_ONLY,
)


@partial(jax.jit, static_argnames=["registered_variables", "config"])
def magnetic_update(
    magnetic_field, gas_state, grid_spacing, dt, registered_variables, config
):
    """Update the magnetic field and gas state.

    Based on a simple fixed point iteration, as done in
    Pang, Dongwen, and Kailiang Wu. "Provably Positivity-Preserving
    Constrained Transport (PPCT) Second-Order Scheme for Ideal
    Magnetohydrodynamics." arXiv preprint arXiv:2410.05173 (2024).
    https://arxiv.org/abs/2410.05173, eq. (3.37).

    Args:
        magnetic_field: The magnetic field.
        gas_state: The gas state.
        grid_spacing: The width of the cells.
        dt: The time step.
        registered_variables: The registered variables.
        config: The simulation configuration.

    Returns:
        The updated magnetic field.
    """

    if config.dimensionality == 2:
        # retrieve the velocity
        # THIS IS ONLY WRITTEN FOR 2D!!!!!
        velocity = jnp.zeros((3, *gas_state.shape[1:]), dtype=magnetic_field.dtype)
        velocity = velocity.at[0:2, :, :].set(
            gas_state[
                registered_variables.velocity_index.x : registered_variables.velocity_index.x
                + 2,
                ...,
            ]
        )
    elif config.dimensionality == 3:
        velocity = gas_state[
            registered_variables.velocity_index.x : registered_variables.velocity_index.x
            + 3,
            ...,
        ]
    else:
        raise ValueError("No 1D MHD.")

    density = gas_state[registered_variables.density_index, ...]

    def phi(velocity, magnetic_field):
        # calculate the electric field, electric field = -v x B
        electric_field = cross(magnetic_field, velocity)

        # calculate the curl of the electric field
        if config.dimensionality == 2:
            curl_electric_field = curl2D(electric_field, grid_spacing)
        elif config.dimensionality == 3:
            curl_electric_field = curl3D(electric_field, grid_spacing)
        else:
            raise ValueError("No 1D curl.")

        # calculate the curl of the magnetic field
        if config.dimensionality == 2:
            curl_magnetic_field = curl2D(magnetic_field, grid_spacing)
        elif config.dimensionality == 3:
            curl_magnetic_field = curl3D(magnetic_field, grid_spacing)
        else:
            raise ValueError("No 1D curl.")

        phi1 = curl_electric_field
        phi2 = cross(magnetic_field, curl_magnetic_field) / density

        return phi1, phi2


    if config.boundary_handling == GHOST_CELLS:
        magnetic_field = _boundary_handler(magnetic_field, config, MAGNETIC_FIELD_ONLY)
        velocity = _boundary_handler(velocity, config, VELOCITY_ONLY)

    B_0 = magnetic_field
    v_0 = velocity

    avg_B = (B_0 + magnetic_field) / 2
    avg_v = (v_0 + velocity) / 2

    phiA, phiB = phi(avg_v, avg_B)

    B_1 = magnetic_field - dt * phiA
    v_1 = velocity - dt * phiB

    if config.boundary_handling == GHOST_CELLS:
        B_1 = _boundary_handler(B_1, config, MAGNETIC_FIELD_ONLY)
        v_1 = _boundary_handler(v_1, config, VELOCITY_ONLY)

    def while_condition(state):
        B_k, v_k, B_kp1, v_kp1, current_iter = state
        max_change = jnp.maximum(
            jnp.max(jnp.linalg.norm(B_k - B_kp1, axis=0)),
            jnp.max(jnp.linalg.norm(v_k - v_kp1, axis=0)),
        )
        
        if config.numerical_precision == SINGLE_PRECISION:
            change_criterion = 1e-5
        elif config.numerical_precision == DOUBLE_PRECISION:
            change_criterion = 1e-10
        else:
            raise ValueError("Unknown numerical precision.")
        
        return (max_change > change_criterion) & (current_iter < 1000)

    def while_body(state):
        B_k, v_k, B_kp1, v_kp1, current_iter = state

        B_k = B_kp1
        v_k = v_kp1

        if config.fv_magnetic_integrator == IMPLICIT_MIDPOINT:
            avg_B = (B_k + magnetic_field) / 2
            avg_v = (v_k + velocity) / 2
        elif config.fv_magnetic_integrator == IMPLICIT_EULER:
            avg_B = B_k
            avg_v = v_k

        phiA, phiB = phi(avg_v, avg_B)

        B_kp1 = magnetic_field - dt * phiA
        v_kp1 = velocity - dt * phiB

        if config.boundary_handling == GHOST_CELLS:
            B_kp1 = _boundary_handler(B_kp1, config, MAGNETIC_FIELD_ONLY)
            v_kp1 = _boundary_handler(v_kp1, config, VELOCITY_ONLY)

        return B_k, v_k, B_kp1, v_kp1, current_iter + 1

    if config.differentiation_mode == FORWARDS:
        _, _, B_n, v_n, _ = jax.lax.while_loop(
            while_condition, while_body, (B_0, v_0, B_1, v_1, 0)
        )
    elif config.differentiation_mode == BACKWARDS:
        _, _, B_n, v_n, _ = checkpointed_while_loop(
            while_condition, while_body, (B_0, v_0, B_1, v_1, 0), checkpoints=3
        )

    if config.runtime_debugging:
        avg_B = (B_n + magnetic_field) / 2
        avg_v = (v_n + velocity) / 2

        phiA, phiB = phi(avg_v, avg_B)

        B_n_mod = magnetic_field - dt * phiA
        v_n_mod = velocity - dt * phiB

        checkify.check(
            jnp.sum(jnp.linalg.norm(B_n - B_n_mod, axis=0)) < 1e-8,
            "Eigeniteration for B not converged!",
        )
        checkify.check(
            jnp.sum(jnp.linalg.norm(v_n - v_n_mod, axis=0)) < 1e-8,
            "Eigeniteration for v not converged!",
        )
        checkify.check(
            jnp.all(
                jnp.abs(
                    divergence2D(B_n, grid_spacing) - divergence2D(B_0, grid_spacing)
                )
                < 1e-8
            ),
            "Divergence of magnetic field not conserved, by {divergence}",
            divergence=jnp.max(
                jnp.abs(
                    divergence2D(B_n, grid_spacing) - divergence2D(B_0, grid_spacing)
                )
            ),
        )

    # update the gas energy
    gas_energy = total_energy_from_primitives(
        density,
        jnp.linalg.norm(velocity, axis=0),
        gas_state[registered_variables.pressure_index, ...],
        5 / 3,
    )
    gas_energy_updated = (
        gas_energy
        - 0.5 * density * jnp.linalg.norm(velocity, axis=0) ** 2
        + 0.5 * density * jnp.linalg.norm(v_n, axis=0) ** 2
    )
    pressure_updated = pressure_from_energy(
        gas_energy_updated, density, jnp.linalg.norm(v_n, axis=0), 5 / 3
    )

    # update the gas state
    if config.dimensionality == 2:
        gas_state = gas_state.at[
            registered_variables.velocity_index.x : registered_variables.velocity_index.x
            + 2,
            ...,
        ].set(v_n[:2, ...])
    elif config.dimensionality == 3:
        gas_state = gas_state.at[
            registered_variables.velocity_index.x : registered_variables.velocity_index.x
            + 3,
            ...,
        ].set(v_n)

    gas_state = gas_state.at[registered_variables.pressure_index, ...].set(
        pressure_updated
    )

    return B_n, gas_state



import jax.numpy as jnp
from astronomix.option_classes.simulation_config import (
    STATE_TYPE,
    STATE_TYPE_ALTERED,
    SimulationConfig,
)


import jax
from beartype import beartype as typechecker
from jaxtyping import jaxtyped


from functools import partial


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config"])
def _unpad(state: STATE_TYPE, config: SimulationConfig) -> STATE_TYPE_ALTERED:
    if config.dimensionality == 1:
        state = jax.lax.slice_in_dim(
            state,
            config.num_ghost_cells,
            state.shape[1] - config.num_ghost_cells,
            axis=1,
        )
    elif config.dimensionality == 2:
        state = jax.lax.slice_in_dim(
            state,
            config.num_ghost_cells,
            state.shape[1] - config.num_ghost_cells,
            axis=1,
        )
        state = jax.lax.slice_in_dim(
            state,
            config.num_ghost_cells,
            state.shape[2] - config.num_ghost_cells,
            axis=2,
        )
    elif config.dimensionality == 3:
        state = jax.lax.slice_in_dim(
            state,
            config.num_ghost_cells,
            state.shape[1] - config.num_ghost_cells,
            axis=1,
        )
        state = jax.lax.slice_in_dim(
            state,
            config.num_ghost_cells,
            state.shape[2] - config.num_ghost_cells,
            axis=2,
        )
        state = jax.lax.slice_in_dim(
            state,
            config.num_ghost_cells,
            state.shape[3] - config.num_ghost_cells,
            axis=3,
        )

    return state


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config"])
def _pad(state: STATE_TYPE, config: SimulationConfig) -> STATE_TYPE_ALTERED:
    if config.dimensionality == 1:
        state = jnp.pad(state, ((0, 0), (2, 2)), mode="edge")

    elif config.dimensionality == 2:
        state = jnp.pad(state, ((0, 0), (2, 2), (2, 2)), mode="edge")

    elif config.dimensionality == 3:
        state = jnp.pad(state, ((0, 0), (2, 2), (2, 2), (2, 2)), mode="edge")

    return state



import shutil


def _show_progress(
    iteration, total, prefix="", suffix="", decimals=1, fill="█", printEnd="\r"
) -> None:
    """
    Progress bar that adapts to terminal width and handles resizing.
    """
    # Get terminal width
    terminal_width = shutil.get_terminal_size((80, 20)).columns

    # Format percentage string
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))

    # Fixed parts (prefix + suffix + percent + " |" + "| " + spaces)
    fixed_part = f"{prefix} | | {percent}% {suffix}"
    fixed_length = len(fixed_part)

    # Compute bar length dynamically
    bar_length = max(10, terminal_width - fixed_length)

    # Compute filled length of the bar
    filledLength = int(bar_length * iteration // total)
    bar = fill * filledLength + "-" * (bar_length - filledLength)

    # Assemble full line
    progress_line = f"{prefix} |{bar}| {percent}% {suffix}"

    # Pad with spaces to ensure full overwrite (avoids leftovers)
    padded_line = progress_line.ljust(terminal_width)

    # Print progress line with carriage return
    print(f"\r{padded_line}", end=printEnd, flush=True)

    # Print newline when complete
    if iteration == total:
        print()



import jax.numpy as jnp
import jax
from functools import partial

from jaxtyping import Array, Float

from typing import Union

from astronomix._finite_difference._fluid_equations._equations import primitive_state_from_conserved_mhd, total_energy_from_primitives_mhd, total_pressure_from_conserved_mhd
from astronomix.variable_registry.registered_variables import AxisInfo, RegisteredVariables
from astronomix.option_classes.simulation_config import (
    FIELD_TYPE,
    STATE_TYPE,
    SimulationConfig,
)

# We only define the flux in x-direction here,
# since the other directions can be obtained
# by permuting the arrays accordingly.
@partial(
    jax.jit, static_argnames=["registered_variables", "config"]
)
def _mhd_flux_x(
    conserved_state: STATE_TYPE,
    minimum_density: Union[float, Float[Array, ""]],
    minimum_pressure: Union[float, Float[Array, ""]],
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:

    primitive_state = primitive_state_from_conserved_mhd(
        conserved_state, minimum_density, minimum_pressure, gamma, config, registered_variables
    )
    
    # retrieve necessary quantities
    rho = primitive_state[registered_variables.density_index]
    v_x = primitive_state[registered_variables.velocity_index.x]
    v_y = primitive_state[registered_variables.velocity_index.y]
    v_z = primitive_state[registered_variables.velocity_index.z]
    B_x = primitive_state[registered_variables.magnetic_index.x]
    B_y = primitive_state[registered_variables.magnetic_index.y]
    B_z = primitive_state[registered_variables.magnetic_index.z]
    p_gas = primitive_state[registered_variables.pressure_index]

    # compute derived quantities
    b_squared = B_x**2 + B_y**2 + B_z**2
    v_squared = v_x**2 + v_y**2 + v_z**2
    total_pressure = p_gas + 0.5 * b_squared
    v_dot_B = v_x * B_x + v_y * B_y + v_z * B_z
    E = total_energy_from_primitives_mhd(
        rho,
        v_squared,
        p_gas,
        b_squared,
        gamma,
    )

    # compute fluxes
    flux = jnp.zeros_like(primitive_state)
    flux = flux.at[registered_variables.density_index].set(rho * v_x)
    flux = flux.at[registered_variables.velocity_index.x].set(rho * v_x**2 + total_pressure - B_x**2)
    flux = flux.at[registered_variables.velocity_index.y].set(rho * v_x * v_y - B_x * B_y)
    flux = flux.at[registered_variables.velocity_index.z].set(rho * v_x * v_z - B_x * B_z)
    flux = flux.at[registered_variables.pressure_index].set((E + total_pressure) * v_x - v_dot_B * B_x)
    flux = flux.at[registered_variables.magnetic_index.x].set(0.0)
    flux = flux.at[registered_variables.magnetic_index.y].set(B_y * v_x - B_x * v_y)
    flux = flux.at[registered_variables.magnetic_index.z].set(B_z * v_x - B_x * v_z)

    return flux


import jax
import jax.numpy as jnp


from functools import partial


@partial(jax.jit, static_argnames=["axis"])
def interp_center_to_face(arr, axis):
    """
    Interpolate to x-interfaces using 4th order
        f_{i+1/2} = (-f_{i-1} + 9f_{i} + 9f_{i+1} - f_{i+2}) / 16
    The i-th array index in the output corresponds to the i+1/2 interface.
    """
    return (
        -jnp.roll(arr, 1, axis=axis)
        + 9 * arr
        + 9 * jnp.roll(arr, -1, axis=axis)
        - jnp.roll(arr, -2, axis=axis)
    ) / 16.0


@partial(jax.jit, static_argnames=["axis"])
def interp_face_to_center(f_int, axis):
    """
    6th order interpolation from face to center.
    """
    return (
        3 * jnp.roll(f_int, 3, axis=axis)
        - 25 * jnp.roll(f_int, 2, axis=axis)
        + 150 * jnp.roll(f_int, 1, axis=axis)
        + 150 * f_int
        - 25 * jnp.roll(f_int, -1, axis=axis)
        + 3 * jnp.roll(f_int, -2, axis=axis)
    ) / 256.0


@partial(jax.jit, static_argnames=["axisA", "axisB"])
def point_values_to_averages(q, axisA, axisB):
    """
    For point values q, we can approximate the cell-averaged
    values Q based on interpolation as

    Q_i = q_i + Δx^2/24 q''(x_i) - ...

    For point values, the second derivative
    can be approximated

    q''(x_i) = (q_{i+1} - 2 q_i + q_{i-1}) / Δx^2

    Here we apply this in two dimensions.
    Compare Buchmüller and Helzel 2014, Eq. 12, 13.

    Such smoothing can be used to retain high-order accuracy
    in dimensionally split settings.    
    """
    smooth_x = (
        jnp.roll(q, 1, axis=axisA) - 2 * q + jnp.roll(q, -1, axis=axisA)
    ) / 24.0
    smooth_y = (
        jnp.roll(q, 1, axis=axisB) - 2 * q + jnp.roll(q, -1, axis=axisB)
    ) / 24.0
    return q + smooth_x + smooth_y


# general
from functools import partial
import jax.numpy as jnp
import jax

# typing
from typing import Tuple, Union
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import (
    CARTESIAN,
    FIELD_TYPE,
    SPHERICAL,
    STATE_TYPE,
    SimulationConfig,
)
from jaxtyping import Array, Int, jaxtyped
from beartype import beartype as typechecker
from typing import Union

from astronomix.option_classes.simulation_params import SimulationParams

# NOTE: currently only works for 1d setups, TODO: generalize


@partial(jax.jit, static_argnames=["config"])
def _calculate_1d_divergence(
    field: FIELD_TYPE, config: SimulationConfig, r: FIELD_TYPE
) -> FIELD_TYPE:
    # calculate the 1d divergence by a simple
    # central difference approximation
    div_field = jnp.zeros_like(field)
    if config.geometry == CARTESIAN:
        div_field = div_field.at[1:-1].set(
            (field[2:] - field[:-2]) / (2 * config.grid_spacing)
        )
    elif config.geometry == SPHERICAL:
        div_field = jnp.zeros_like(field)
        # this is not exactly correct, as our field values are
        # defined at the volumetric not geometric cell centers etc
        # but should be fine for the shock finder
        div_field = div_field.at[1:-1].set(
            (r[2:] ** 2 * field[2:] - r[:-2] ** 2 * field[:-2])
            / (2 * config.grid_spacing * r[1:-1] ** 2)
        )
    else:
        raise NotImplementedError(
            "Only Cartesian and Spherical geometry supported for the shock finder."
        )
    return div_field


@jax.jit
def shock_sensor(pressure: FIELD_TYPE) -> FIELD_TYPE:
    """
    WENO-JS 1D smoothness indicator for shock detection.

    Args:
        pressure: the 1d pressure

    Returns:
        shock sensors, high where large pressure jumps

    """

    shock_sensors = jnp.zeros_like(pressure)
    shock_sensors = shock_sensors.at[1:-1].set(
        1 / 4 * (pressure[2:] - pressure[:-2]) ** 2
        + 13 / 12 * (pressure[2:] - 2 * pressure[1:-1] + pressure[:-2])
    )

    return shock_sensors


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["registered_variables", "config"])
def shock_criteria(
    primitive_state: STATE_TYPE,
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    helper_data: HelperData,
) -> jnp.ndarray:
    """
    Implement the shock criteria from Pfrommer et al, 2017.
    https://arxiv.org/abs/1604.07399

    # NOTE: for now only 1D

    """

    gamma_gas = 5 / 3
    gamma_cr = 4 / 3

    # get the velocity
    velocity = primitive_state[registered_variables.velocity_index]

    # get the cosmic ray pressure
    P_CRs = primitive_state[registered_variables.cosmic_ray_n_index] ** gamma_cr

    # i) \nabla \cdot \vec{v} < 0
    div_v = _calculate_1d_divergence(velocity, config, helper_data.geometric_centers)
    converging_flow_criterion = div_v < 0

    # ii) \nabla T \cdot \nabla \rho > 0
    pseudo_temperature = (
        primitive_state[registered_variables.pressure_index]
        / primitive_state[registered_variables.density_index]
    )
    div_T = jnp.zeros_like(pseudo_temperature)
    div_T = div_T.at[1:-1].set((pseudo_temperature[2:] - pseudo_temperature[:-2]) / 2)
    div_rho = jnp.zeros_like(primitive_state[registered_variables.density_index])
    div_rho = div_rho.at[1:-1].set(
        (
            primitive_state[registered_variables.density_index][2:]
            - primitive_state[registered_variables.density_index][:-2]
        )
        / 2
    )
    no_spurious_shocks = div_T * div_rho > 0

    # iii) M1 > Mmin
    Mmin = 1.3
    # NOTE: currently we only consider shocks moving left to right
    P2 = primitive_state[registered_variables.pressure_index, :-2]
    P2_CRs = P_CRs[:-2]
    e2_gas = P2 - P2_CRs  # gas energy / volume
    e2_crs = P2_CRs / (gamma_cr - 1)  # cosmic ray energy / volume
    e2 = e2_gas + e2_crs  # total energy / volume
    rho2 = primitive_state[registered_variables.density_index, :-2]

    P1 = primitive_state[registered_variables.pressure_index, 2:]
    P1_CRs = P_CRs[2:]
    P1_gas = P1 - P1_CRs  # gas pressure
    e1_gas = P1_gas / (gamma_gas - 1)  # gas energy / volume
    e1_crs = P1_CRs / (gamma_cr - 1)  # cosmic ray energy / volume
    e1 = e1_gas + e1_crs  # total energy / volume
    rho1 = primitive_state[registered_variables.density_index, 2:]

    gamma_eff1 = (gamma_cr * P1_CRs + gamma_gas * P1_gas) / P1
    gamma_eff2 = (gamma_cr * P2_CRs + gamma_gas * P2) / P2

    gamma1 = P1 / e1 + 1
    gamma2 = P2 / e2 + 1

    gammat = P2 / P1

    C = ((gamma2 + 1) * gammat + gamma2 - 1) * (gamma1 - 1)

    # advanced Mach number calculation, formula 16 from Dubois et al, 2019
    denominator = jnp.where(
        jnp.abs(C - ((gamma1 + 1) + (gamma1 - 1) * gammat) * (gamma2 - 1)) > 1e-6,
        (C - ((gamma1 + 1) + (gamma1 - 1) * gammat) * (gamma2 - 1)),
        1e-6,
    )
    M1sq = 1 / gamma_eff2 * (gammat - 1) * C / denominator

    # simple Mach number calculation, crashes
    # the simulation where x_s = 1, better just evaluate
    # this where the other criterions hold / add a numerical
    # safeguard
    # x_s = rho2 / rho1
    # M1sq = (P2 / P1 - 1) * x_s / (gamma_eff1 * (x_s - 1))

    mach_number_criterion = jnp.zeros_like(converging_flow_criterion, dtype=jnp.bool_)

    mach_number_criterion = mach_number_criterion.at[1:-1].set(M1sq > Mmin**2)

    return converging_flow_criterion & no_spurious_shocks & mach_number_criterion


@partial(jax.jit, static_argnames=["registered_variables", "config"])
def find_shock_zone(
    primitive_state: STATE_TYPE,
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    helper_data: HelperData,
) -> Tuple[
    Union[int, Int[Array, ""]], Union[int, Int[Array, ""]], Union[int, Int[Array, ""]]
]:
    """
    Find a numerically broadened shock region based of the strongest shock based
    on the result of the shock_sensor function and the pressure difference
    between adjacent cells. Assumes a shock front moving left to right.

    Args:
        pressure: 1d pressure
        velocity: 1d velocity

    Returns:
        index of max shock sensor,
        left boundary of broadened shock,
        right boundary of broadened shock

    """

    pressure = primitive_state[registered_variables.pressure_index]
    num_cells = pressure.shape[0]

    # one can either use the maximum of the shock sensor
    sensors = shock_sensor(pressure)
    # or the cell with maximum compression, as in Pfrommer et al 2017
    # div_v = _calculate_1d_divergence(primitive_state[registered_variables.velocity_index], config, helper_data.geometric_centers)

    shock_crit = shock_criteria(
        primitive_state, config, registered_variables, helper_data
    )

    max_shock_idx = jnp.argmax(jnp.where(shock_crit, sensors, -1))
    # max_shock_idx = jnp.argmin(jnp.where(shock_crit, div_v, 1))

    # calculate differences in pressure
    pressure_differences = jnp.zeros_like(pressure)
    # 0 <- 1 - 0
    pressure_differences = pressure_differences.at[1:].set(pressure[1:] - pressure[:-1])

    # bound on the change in pressure between adjacent cells compared
    # to the pressure jump at the max_shock_index
    bound_diff = 0.1 * jnp.abs(pressure_differences[max_shock_idx])

    # left index: closest left index where |pressure_difference| < bound_diff or switched sign
    # right index: closest right index where |pressure_difference| < bound_diff or switched sign
    indices = jnp.arange(num_cells)
    left_indices = jnp.where(
        (indices < max_shock_idx)
        & ((jnp.abs(pressure_differences) < bound_diff) | (pressure_differences > 0)),
        indices,
        -1,
    )
    right_indices = jnp.where(
        (indices > max_shock_idx)
        & ((jnp.abs(pressure_differences) < bound_diff) | (pressure_differences < 0)),
        indices,
        num_cells,
    )
    left_idx = jnp.max(left_indices)
    right_idx = jnp.min(right_indices)

    return max_shock_idx, left_idx, right_idx



# general
import jax
import jax.numpy as jnp
from functools import partial

# type checking
from beartype import beartype as typechecker
from jaxtyping import Array, Float, jaxtyped
from typing import Union

# astronomix containers
from astronomix._physics_modules._cosmic_rays.cr_fluid_equations import (
    total_energy_from_primitives_with_crs,
)
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import STATE_TYPE, SimulationConfig
from astronomix.data_classes.simulation_helper_data import HelperData

# astronomix functions
from astronomix._physics_modules._self_gravity._poisson_solver import (
    _compute_gravitational_potential,
)
from astronomix._fluid_equations._equations import (
    get_absolute_velocity,
    total_energy_from_primitives,
)


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def calculate_internal_energy(state, helper_data, gamma, config, registered_variables):
    p = state[registered_variables.pressure_index]

    if config.cosmic_ray_config.cosmic_rays:
        gamma_cr = 4 / 3
        p = p - state[registered_variables.cosmic_ray_n_index] ** gamma_cr

    internal_energy = p / (gamma - 1)

    if config.dimensionality == 1:
        return jnp.sum(internal_energy * helper_data.cell_volumes)
    else:
        return jnp.sum(internal_energy * config.grid_spacing**config.dimensionality)


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def calculate_radial_momentum(state, helper_data, config, registered_variables):
    rho = state[registered_variables.density_index]
    box_center = jnp.zeros(config.dimensionality) + config.box_size / 2
    geometric_centers = helper_data.geometric_centers
    r_hat = (geometric_centers - box_center) / jnp.linalg.norm(
        geometric_centers - box_center, axis=-1, keepdims=True
    )

    if config.dimensionality == 1:
        u = state[registered_variables.velocity_index]
    else:
        u = state[
            registered_variables.velocity_index.x : registered_variables.velocity_index.x
            + config.dimensionality
        ]

    u_radial = jnp.sum(jnp.moveaxis(u, 0, -1) * r_hat, axis=-1)

    radial_momentum = rho * u_radial

    if config.dimensionality == 1:
        return jnp.sum(radial_momentum * helper_data.cell_volumes)
    else:
        return jnp.sum(radial_momentum * config.grid_spacing**config.dimensionality)


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def calculate_kinetic_energy(state, helper_data, config, registered_variables):
    rho = state[registered_variables.density_index]
    u = get_absolute_velocity(state, config, registered_variables)

    kinetic_energy = 0.5 * rho * u**2

    if config.dimensionality == 1:
        return jnp.sum(kinetic_energy * helper_data.cell_volumes)
    else:
        return jnp.sum(kinetic_energy * config.grid_spacing**config.dimensionality)


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def calculate_gravitational_energy(
    state, helper_data, gravitational_constant, config, registered_variables
):
    rho = state[registered_variables.density_index]

    potential = _compute_gravitational_potential(
        rho, config.grid_spacing, config, gravitational_constant
    )
    gravitational_energy = 0.5 * rho * potential
    if config.dimensionality == 1:
        return jnp.sum(gravitational_energy * helper_data.cell_volumes)
    else:
        return jnp.sum(
            gravitational_energy * config.grid_spacing**config.dimensionality
        )


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def calculate_total_energy(
    primitive_state: STATE_TYPE,
    helper_data: HelperData,
    gamma: Union[float, Float[Array, ""]],
    gravitational_constant: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
) -> Float[Array, ""]:
    """
    Calculate the total energy in the domain.

    Args:
        primitive_state: The primitive state array.
        helper_data: The helper data.
        gamma: The adiabatic index.
        num_ghost_cells: The number of ghost cells.

    Returns:
        The total energy.
    """

    rho = primitive_state[registered_variables.density_index]
    u = get_absolute_velocity(primitive_state, config, registered_variables)
    p = primitive_state[registered_variables.pressure_index]

    if config.cosmic_ray_config.cosmic_rays:
        energy = total_energy_from_primitives_with_crs(
            primitive_state, registered_variables
        )
    else:
        energy = total_energy_from_primitives(rho, u, p, gamma)

    if config.self_gravity:
        potential = _compute_gravitational_potential(
            rho, config.grid_spacing, config, gravitational_constant
        )
        energy += 0.5 * rho * potential

    if config.dimensionality == 1:
        return jnp.sum(energy * helper_data.cell_volumes)
    else:
        return jnp.sum(energy * config.grid_spacing**config.dimensionality)


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config"])
def calculate_total_mass(
    primitive_state: STATE_TYPE,
    helper_data: HelperData,
    config: SimulationConfig,
) -> Float[Array, ""]:
    """
    Calculate the total mass in the domain.

    Args:
        primitive_state: The primitive state array.
        helper_data: The helper data.
        config: The simulation configuration.

    Returns:
        The total mass.
    """
    num_ghost_cells = config.num_ghost_cells

    if config.dimensionality == 1:
        return jnp.sum(primitive_state[0] * helper_data.cell_volumes)
    else:
        return jnp.sum(primitive_state[0]) * config.box_size**config.dimensionality



from typing import Union
import jax
from functools import partial
import jax.numpy as jnp

from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker

from astronomix._physics_modules._cnn_mhd_corrector._cnn_mhd_corrector import (
    _cnn_mhd_corrector,
)
from astronomix._physics_modules._cooling._cooling import first_order_pressure_update, update_pressure_by_cooling
from astronomix._physics_modules._cosmic_rays.cr_injection import (
    inject_crs_at_strongest_shock,
)
from astronomix._physics_modules._neural_net_force._neural_net_force import (
    _neural_net_force,
)
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix._geometry.boundaries import _boundary_handler
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import (
    SPHERICAL,
    STATE_TYPE,
    SimulationConfig,
)
from astronomix.option_classes.simulation_params import SimulationParams
from astronomix._physics_modules._stellar_wind.stellar_wind import _wind_injection
from astronomix.shock_finder.shock_finder import shock_criteria


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _run_physics_modules(
    primitive_state: STATE_TYPE,
    dt: Float[Array, ""],
    config: SimulationConfig,
    params: SimulationParams,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
    current_time: Union[float, Float[Array, ""]],
) -> STATE_TYPE:
    """Run all the physics modules. The physics modules are switched on/off and
    configured in the simulation configuration. Parameters for the physics modules
    (with respect to which the simulation can be differentiated) are stored in the
    simulation parameters.

    Args:
        primitive_state: The primitive state array.
        dt: The time step.
        config: The simulation configuration.
        params: The simulation parameters.
        helper_data: The helper data.

    Returns:
        The primitive state array with the physics modules applied.
    """

    # stellar wind
    if config.wind_config.stellar_wind:
        primitive_state = _wind_injection(
            primitive_state, dt, config, params, helper_data, registered_variables
        )

        # we might want to run the boundary handler after all physics modules have completed
        # primitive_state = _boundary_handler(primitive_state, config.left_boundary, config.right_boundary)

    if config.cosmic_ray_config.diffusive_shock_acceleration:
        shock_crit = shock_criteria(
            primitive_state, config, registered_variables, helper_data
        )

        # injecting cosmic rays only after a certain amount of time
        # is an ad-hoc fix to problems that come about when a shock
        # has not yet properly formed
        primitive_state = jax.lax.cond(
            jnp.logical_and(
                current_time
                >= params.cosmic_ray_params.diffusive_shock_acceleration_start_time,
                jnp.any(shock_crit),
            ),
            lambda primitive_state: inject_crs_at_strongest_shock(
                primitive_state,
                params.gamma,
                helper_data,
                params.cosmic_ray_params,
                config,
                registered_variables,
                dt,
            ),
            lambda primitive_state: primitive_state,
            primitive_state,
        )

    if config.cooling_config.cooling:
        primitive_state = update_pressure_by_cooling(
            primitive_state,
            registered_variables,
            config.cooling_config,
            params,
            dt,
        )
        # primitive_state = first_order_pressure_update(
        #     primitive_state,
        #     registered_variables,
        #     config,
        #     helper_data,
        #     params,
        #     dt
        # )

    if config.neural_net_force_config.neural_net_force:
        primitive_state = _neural_net_force(
            primitive_state,
            config,
            registered_variables,
            params,
            helper_data,
            dt,
            current_time,
        )

    if config.cnn_mhd_corrector_config.cnn_mhd_corrector:
        primitive_state = _cnn_mhd_corrector(
            primitive_state, config, registered_variables, params, dt
        )

    return primitive_state
